# ruff: noqa: UP006 UP007
# @omlish-lite
# Based on bluelet ( https://github.com/sampsyo/bluelet ) by Adrian Sampson, original license:
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import abc
import dataclasses as dc
import time
import types
import typing as ta

from .events import BlueletEvent
from .events import BlueletFuture
from .events import WaitableBlueletEvent


T = ta.TypeVar('T')

BlueletExcInfo = ta.Tuple[ta.Type[BaseException], BaseException, types.TracebackType]  # ta.TypeAlias

BlueletCoro = ta.Generator[ta.Union['BlueletEvent', 'BlueletCoro'], ta.Any, None]  # ta.TypeAlias

BlueletSpawnable = ta.Union[BlueletCoro, ta.Awaitable]  # ta.TypeAlias


##


@dc.dataclass(frozen=True, eq=False)
class _BlueletAwaitableDriver:
    a: ta.Awaitable

    def __call__(self) -> BlueletCoro:
        g = self.a.__await__()
        gi = iter(g)
        while True:
            try:
                f = gi.send(None)
            except StopIteration as e:
                yield ReturnBlueletEvent(e.value)
                break
            else:
                if not isinstance(f, BlueletFuture):
                    raise TypeError(f)
                res = yield f.event
                f.done = True
                f.result = res


##


class CoreBlueletEvent(BlueletEvent, abc.ABC):  # noqa
    pass


@dc.dataclass(frozen=True, eq=False)
class ValueBlueletEvent(CoreBlueletEvent, ta.Generic[T]):
    """An event that does nothing but return a fixed value."""

    value: T


@dc.dataclass(frozen=True, eq=False)
class ExceptionBlueletEvent(CoreBlueletEvent):
    """Raise an exception at the yield point. Used internally."""

    exc_info: BlueletExcInfo


@dc.dataclass(frozen=True, eq=False)
class SpawnBlueletEvent(CoreBlueletEvent):
    """Add a new coroutine coro to the scheduler."""

    spawned: BlueletSpawnable


@dc.dataclass(frozen=True, eq=False)
class JoinBlueletEvent(CoreBlueletEvent):
    """Suspend the coro until the specified child coro has completed."""

    child: BlueletCoro


@dc.dataclass(frozen=True, eq=False)
class KillBlueletEvent(CoreBlueletEvent):
    """Unschedule a child coro."""

    child: BlueletCoro


@dc.dataclass(frozen=True, eq=False)
class DelegationBlueletEvent(CoreBlueletEvent):
    """
    Suspend execution of the current coro, start a new coro and, once the child coro finished, return control to
    the parent coro.
    """

    spawned: BlueletCoro


@dc.dataclass(frozen=True, eq=False)
class ReturnBlueletEvent(CoreBlueletEvent, ta.Generic[T]):
    """Return a value the current coro's delegator at the point of delegation. Ends the current (delegate) coro."""

    value: ta.Optional[T]


@dc.dataclass(frozen=True, eq=False)
class SleepBlueletEvent(WaitableBlueletEvent, CoreBlueletEvent):
    """Suspend the coro for a given duration."""

    wakeup_time: float

    def time_left(self) -> float:
        return max(self.wakeup_time - time.time(), 0.)


##


class _CoreBlueletApi:
    def value(self, v: T) -> ValueBlueletEvent[T]:
        """Event: yield a value."""

        return ValueBlueletEvent(v)

    def null(self) -> ValueBlueletEvent[None]:
        """Event: yield to the scheduler without doing anything special."""

        return ValueBlueletEvent(None)

    def spawn(self, spawned: BlueletSpawnable) -> SpawnBlueletEvent:
        """Event: add another coroutine to the scheduler. Both the parent and child coroutines run concurrently."""

        if isinstance(spawned, types.CoroutineType):
            spawned = _BlueletAwaitableDriver(spawned)()

        if not isinstance(spawned, types.GeneratorType):
            raise TypeError(f'{spawned} is not spawnable')

        return SpawnBlueletEvent(spawned)

    def join(self, coro: BlueletCoro) -> JoinBlueletEvent:
        """Suspend the coro until another, previously `spawn`ed coro completes."""

        return JoinBlueletEvent(coro)

    def kill(self, coro: BlueletCoro) -> KillBlueletEvent:
        """Halt the execution of a different `spawn`ed coro."""

        return KillBlueletEvent(coro)

    def call(self, spawned: BlueletSpawnable) -> DelegationBlueletEvent:
        """
        Event: delegate to another coroutine. The current coroutine is resumed once the sub-coroutine finishes. If the
        sub-coroutine returns a value using end(), then this event returns that value.
        """

        if isinstance(spawned, types.CoroutineType):
            spawned = _BlueletAwaitableDriver(spawned)()

        if not isinstance(spawned, types.GeneratorType):
            raise TypeError(f'{spawned} is not spawnable')

        return DelegationBlueletEvent(spawned)

    def end(self, value: ta.Optional[T] = None) -> ReturnBlueletEvent[T]:
        """Event: ends the coroutine and returns a value to its delegator."""

        return ReturnBlueletEvent(value)

    def sleep(self, duration: float) -> SleepBlueletEvent:
        """Event: suspend the coro for ``duration`` seconds."""

        return SleepBlueletEvent(time.time() + duration)



# @omlish-lite
# ruff: noqa: UP006 UP007
import abc
import contextlib
import logging
import selectors
import threading
import typing as ta

from ..addresses import SocketAndAddress
from ..bind import SocketBinder
from ..io import close_socket_immediately
from .handlers import SocketServerHandler


##


class SocketServer(abc.ABC):
    _DEFAULT_LOGGER = logging.getLogger('.'.join([__name__, 'SocketServer']))

    def __init__(
            self,
            binder: SocketBinder,
            handler: SocketServerHandler,
            *,
            on_error: ta.Optional[ta.Callable[[BaseException, ta.Optional[SocketAndAddress]], None]] = None,
            error_logger: ta.Optional[logging.Logger] = _DEFAULT_LOGGER,
            poll_interval: float = .5,
            shutdown_timeout: ta.Optional[float] = None,
    ) -> None:
        super().__init__()

        self._binder = binder
        self._handler = handler
        self._on_error = on_error
        self._error_logger = error_logger
        self._poll_interval = poll_interval
        self._shutdown_timeout = shutdown_timeout

        self._lock = threading.RLock()
        self._is_shutdown = threading.Event()
        self._should_shutdown = False

    @property
    def binder(self) -> SocketBinder:
        return self._binder

    @property
    def handler(self) -> SocketServerHandler:
        return self._handler

    #

    def _handle_error(self, exc: BaseException, conn: ta.Optional[SocketAndAddress] = None) -> None:
        if (error_logger := self._error_logger) is not None:
            error_logger.exception('Error in socket server: %r', conn)

        if (on_error := self._on_error) is not None:
            on_error(exc, conn)

    #

    class SelectorProtocol(ta.Protocol):
        def register(self, *args, **kwargs) -> None:
            raise NotImplementedError

        def select(self, *args, **kwargs) -> bool:
            raise NotImplementedError

    Selector: ta.ClassVar[ta.Any]
    if hasattr(selectors, 'PollSelector'):
        Selector = selectors.PollSelector
    else:
        Selector = selectors.SelectSelector

    #

    @contextlib.contextmanager
    def _listen_context(self) -> ta.Iterator[SelectorProtocol]:
        with contextlib.ExitStack() as es:
            es.enter_context(self._lock)
            es.enter_context(self._binder)

            self._binder.listen()

            self._is_shutdown.clear()
            try:
                # XXX: Consider using another file descriptor or connecting to the socket to wake this up instead of
                # polling. Polling reduces our responsiveness to a shutdown request and wastes cpu at all other times.
                with self.Selector() as selector:
                    selector.register(self._binder.fileno(), selectors.EVENT_READ)

                    yield selector

            finally:
                self._is_shutdown.set()

    @contextlib.contextmanager
    def loop_context(self, poll_interval: ta.Optional[float] = None) -> ta.Iterator[ta.Iterator[bool]]:
        if poll_interval is None:
            poll_interval = self._poll_interval

        with self._listen_context() as selector:
            def loop():
                while not self._should_shutdown:
                    ready = selector.select(poll_interval)

                    # bpo-35017: shutdown() called during select(), exit immediately.
                    if self._should_shutdown:
                        break  # type: ignore[unreachable]

                    if ready:
                        try:
                            conn = self._binder.accept()

                        except OSError as exc:
                            self._handle_error(exc)

                            return

                        try:
                            self._handler(conn)

                        except Exception as exc:  # noqa
                            self._handle_error(exc, conn)

                            close_socket_immediately(conn.socket)

                    yield bool(ready)

            yield loop()

    def run(self, poll_interval: ta.Optional[float] = None) -> None:
        with self.loop_context(poll_interval=poll_interval) as loop:
            for _ in loop:
                pass

    #

    class _NOT_SET:  # noqa
        def __new__(cls, *args, **kwargs):  # noqa
            raise TypeError

    def shutdown(
            self,
            block: bool = False,
            timeout: ta.Union[float, None, ta.Type[_NOT_SET]] = _NOT_SET,
    ) -> None:
        self._should_shutdown = True

        if block:
            if timeout is self._NOT_SET:
                timeout = self._shutdown_timeout

            if not self._is_shutdown.wait(timeout=timeout):  # type: ignore
                raise TimeoutError

    #

    def __enter__(self) -> 'SocketServer':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown()



# ruff: noqa: UP006 UP007
"""
TODO:
 - Event (/ Predicate)
"""
import abc
import time
import typing as ta


TimeoutLike = ta.Union['Timeout', 'Timeout.Default', ta.Iterable['TimeoutLike'], float]  # ta.TypeAlias


##


class Timeout(abc.ABC):
    @property
    @abc.abstractmethod
    def can_expire(self) -> bool:
        """Indicates whether or not this timeout will ever expire."""

        raise NotImplementedError

    @abc.abstractmethod
    def expired(self) -> bool:
        """Return whether or not this timeout has expired."""

        raise NotImplementedError

    @abc.abstractmethod
    def remaining(self) -> float:
        """Returns the time (in seconds) remaining until the timeout expires. May be negative and/or infinite."""

        raise NotImplementedError

    @abc.abstractmethod
    def __call__(self) -> float:
        """Returns the time (in seconds) remaining until the timeout expires, or raises if the timeout has expired."""

        raise NotImplementedError

    @abc.abstractmethod
    def or_(self, o: ta.Any) -> ta.Any:
        """Evaluates time remaining via remaining() if this timeout can expire, otherwise returns `o`."""

        raise NotImplementedError

    #

    @classmethod
    def _now(cls) -> float:
        return time.time()

    #

    class Default:
        def __new__(cls, *args, **kwargs):  # noqa
            raise TypeError

    class _NOT_SPECIFIED:  # noqa
        def __new__(cls, *args, **kwargs):  # noqa
            raise TypeError

    @classmethod
    def of(
            cls,
            obj: ta.Optional[TimeoutLike],
            default: ta.Union[TimeoutLike, ta.Type[_NOT_SPECIFIED]] = _NOT_SPECIFIED,
    ) -> 'Timeout':
        if obj is None:
            return InfiniteTimeout()

        elif isinstance(obj, Timeout):
            return obj

        elif isinstance(obj, (float, int)):
            return DeadlineTimeout(cls._now() + obj)

        elif isinstance(obj, ta.Iterable):
            return CompositeTimeout(*[Timeout.of(c) for c in obj])

        elif obj is Timeout.Default:
            if default is Timeout._NOT_SPECIFIED or default is Timeout.Default:
                raise RuntimeError('Must specify a default timeout')

            else:
                return Timeout.of(default)  # type: ignore[arg-type]

        else:
            raise TypeError(obj)

    @classmethod
    def of_deadline(cls, deadline: float) -> 'DeadlineTimeout':
        return DeadlineTimeout(deadline)

    @classmethod
    def of_predicate(cls, expired_fn: ta.Callable[[], bool]) -> 'PredicateTimeout':
        return PredicateTimeout(expired_fn)


class DeadlineTimeout(Timeout):
    def __init__(
            self,
            deadline: float,
            exc: ta.Union[ta.Type[BaseException], BaseException] = TimeoutError,
    ) -> None:
        super().__init__()

        self.deadline = deadline
        self.exc = exc

    @property
    def can_expire(self) -> bool:
        return True

    def expired(self) -> bool:
        return not (self.remaining() > 0)

    def remaining(self) -> float:
        return self.deadline - self._now()

    def __call__(self) -> float:
        if (rem := self.remaining()) > 0:
            return rem
        raise self.exc

    def or_(self, o: ta.Any) -> ta.Any:
        return self()


class InfiniteTimeout(Timeout):
    @property
    def can_expire(self) -> bool:
        return False

    def expired(self) -> bool:
        return False

    def remaining(self) -> float:
        return float('inf')

    def __call__(self) -> float:
        return float('inf')

    def or_(self, o: ta.Any) -> ta.Any:
        return o


class CompositeTimeout(Timeout):
    def __init__(self, *children: Timeout) -> None:
        super().__init__()

        self.children = children

    @property
    def can_expire(self) -> bool:
        return any(c.can_expire for c in self.children)

    def expired(self) -> bool:
        return any(c.expired() for c in self.children)

    def remaining(self) -> float:
        return min(c.remaining() for c in self.children)

    def __call__(self) -> float:
        return min(c() for c in self.children)

    def or_(self, o: ta.Any) -> ta.Any:
        if self.can_expire:
            return self()
        return o


class PredicateTimeout(Timeout):
    def __init__(
            self,
            expired_fn: ta.Callable[[], bool],
            exc: ta.Union[ta.Type[BaseException], BaseException] = TimeoutError,
    ) -> None:
        super().__init__()

        self.expired_fn = expired_fn
        self.exc = exc

    @property
    def can_expire(self) -> bool:
        return True

    def expired(self) -> bool:
        return self.expired_fn()

    def remaining(self) -> float:
        return float('inf')

    def __call__(self) -> float:
        if not self.expired_fn():
            return float('inf')
        raise self.exc

    def or_(self, o: ta.Any) -> ta.Any:
        return self()



# ruff: noqa: UP006 UP007
# @omlish-lite
import subprocess
import typing as ta


##


def subprocess_close(
        proc: subprocess.Popen,
        timeout: ta.Optional[float] = None,
) -> None:
    # TODO: terminate, sleep, kill
    if proc.stdout:
        proc.stdout.close()
    if proc.stderr:
        proc.stderr.close()
    if proc.stdin:
        proc.stdin.close()

    proc.wait(timeout)



# ruff: noqa: UP006 UP007
# @omlish-lite
# PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
# --------------------------------------------
#
# 1. This LICENSE AGREEMENT is between the Python Software Foundation ("PSF"), and the Individual or Organization
# ("Licensee") accessing and otherwise using this software ("Python") in source or binary form and its associated
# documentation.
#
# 2. Subject to the terms and conditions of this License Agreement, PSF hereby grants Licensee a nonexclusive,
# royalty-free, world-wide license to reproduce, analyze, test, perform and/or display publicly, prepare derivative
# works, distribute, and otherwise use Python alone or in any derivative version, provided, however, that PSF's License
# Agreement and PSF's notice of copyright, i.e., "Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
# 2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017 Python Software Foundation; All Rights Reserved" are retained in Python
# alone or in any derivative version prepared by Licensee.
#
# 3. In the event Licensee prepares a derivative work that is based on or incorporates Python or any part thereof, and
# wants to make the derivative work available to others as provided herein, then Licensee hereby agrees to include in
# any such work a brief summary of the changes made to Python.
#
# 4. PSF is making Python available to Licensee on an "AS IS" basis.  PSF MAKES NO REPRESENTATIONS OR WARRANTIES,
# EXPRESS OR IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND DISCLAIMS ANY REPRESENTATION OR WARRANTY
# OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF PYTHON WILL NOT INFRINGE ANY THIRD PARTY
# RIGHTS.
#
# 5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL
# DAMAGES OR LOSS AS A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON, OR ANY DERIVATIVE THEREOF, EVEN IF
# ADVISED OF THE POSSIBILITY THEREOF.
#
# 6. This License Agreement will automatically terminate upon a material breach of its terms and conditions.
#
# 7. Nothing in this License Agreement shall be deemed to create any relationship of agency, partnership, or joint
# venture between PSF and Licensee.  This License Agreement does not grant permission to use PSF trademarks or trade
# name in a trademark sense to endorse or promote products or services of Licensee, or any third party.
#
# 8. By copying, installing or otherwise using Python, Licensee agrees to be bound by the terms and conditions of this
# License Agreement.
"""
"Test suite" lol:

curl -v localhost:8000
curl -v localhost:8000 -d 'foo'
curl -v -XFOO localhost:8000 -d 'foo'
curl -v -XPOST -H 'Expect: 100-Continue' localhost:8000 -d 'foo'

curl -v -0 localhost:8000
curl -v -0 localhost:8000 -d 'foo'
curl -v -0 -XFOO localhost:8000 -d 'foo'

curl -v -XPOST localhost:8000 -d 'foo' --next -XPOST localhost:8000 -d 'bar'
curl -v -XPOST localhost:8000 -d 'foo' --next -XFOO localhost:8000 -d 'bar'
curl -v -XFOO localhost:8000 -d 'foo' --next -XPOST localhost:8000 -d 'bar'
curl -v -XFOO localhost:8000 -d 'foo' --next -XFOO localhost:8000 -d 'bar'
"""
import abc
import dataclasses as dc
import email.utils
import html
import http.client
import http.server
import io
import textwrap
import time
import typing as ta

from ...lite.check import check
from ...sockets.addresses import SocketAddress
from ..handlers import HttpHandler
from ..handlers import HttpHandlerRequest
from ..handlers import HttpHandlerResponseData
from ..handlers import HttpHandlerResponseStreamedData
from ..handlers import UnsupportedMethodHttpHandlerError
from ..parsing import EmptyParsedHttpResult
from ..parsing import HttpRequestParser
from ..parsing import ParsedHttpRequest
from ..parsing import ParseHttpRequestError
from ..versions import HttpProtocolVersion
from ..versions import HttpProtocolVersions


CoroHttpServerFactory = ta.Callable[[SocketAddress], 'CoroHttpServer']


##


class CoroHttpServer:
    """
    Adapted from stdlib:
     - https://github.com/python/cpython/blob/4b4e0dbdf49adc91c35a357ad332ab3abd4c31b1/Lib/http/server.py#L146
    """

    #

    def __init__(
            self,
            client_address: SocketAddress,
            *,
            handler: HttpHandler,
            parser: HttpRequestParser = HttpRequestParser(),

            default_content_type: ta.Optional[str] = None,

            error_message_format: ta.Optional[str] = None,
            error_content_type: ta.Optional[str] = None,
    ) -> None:
        super().__init__()

        self._client_address = client_address

        self._handler = handler
        self._parser = parser

        self._default_content_type = default_content_type or self.DEFAULT_CONTENT_TYPE

        self._error_message_format = error_message_format or self.DEFAULT_ERROR_MESSAGE
        self._error_content_type = error_content_type or self.DEFAULT_ERROR_CONTENT_TYPE

    #

    @property
    def client_address(self) -> SocketAddress:
        return self._client_address

    @property
    def handler(self) -> HttpHandler:
        return self._handler

    @property
    def parser(self) -> HttpRequestParser:
        return self._parser

    #

    def _format_timestamp(self, timestamp: ta.Optional[float] = None) -> str:
        if timestamp is None:
            timestamp = time.time()
        return email.utils.formatdate(timestamp, usegmt=True)

    #

    def _header_encode(self, s: str) -> bytes:
        return s.encode('latin-1', 'strict')

    class _Header(ta.NamedTuple):
        key: str
        value: str

    def _format_header_line(self, h: _Header) -> str:
        return f'{h.key}: {h.value}\r\n'

    def _get_header_close_connection_action(self, h: _Header) -> ta.Optional[bool]:
        if h.key.lower() != 'connection':
            return None
        elif h.value.lower() == 'close':
            return True
        elif h.value.lower() == 'keep-alive':
            return False
        else:
            return None

    def _make_default_headers(self) -> ta.List[_Header]:
        return [
            self._Header('Date', self._format_timestamp()),
        ]

    #

    _STATUS_RESPONSES: ta.Mapping[int, ta.Tuple[str, str]] = {
        v: (v.phrase, v.description)
        for v in http.HTTPStatus.__members__.values()
    }

    def _format_status_line(
            self,
            version: HttpProtocolVersion,
            code: ta.Union[http.HTTPStatus, int],
            message: ta.Optional[str] = None,
    ) -> str:
        if message is None:
            if code in self._STATUS_RESPONSES:
                message = self._STATUS_RESPONSES[code][0]
            else:
                message = ''

        return f'{version} {int(code)} {message}\r\n'

    #

    @dc.dataclass(frozen=True)
    class _Response:
        version: HttpProtocolVersion
        code: http.HTTPStatus

        message: ta.Optional[str] = None
        headers: ta.Optional[ta.Sequence['CoroHttpServer._Header']] = None
        data: ta.Optional[HttpHandlerResponseData] = None
        close_connection: ta.Optional[bool] = False

        def get_header(self, key: str) -> ta.Optional['CoroHttpServer._Header']:
            for h in self.headers or []:
                if h.key.lower() == key.lower():
                    return h
            return None

        def close(self) -> None:
            if isinstance(d := self.data, HttpHandlerResponseStreamedData):
                d.close()

    #

    def _build_response_head_bytes(self, a: _Response) -> bytes:
        out = io.BytesIO()

        if a.version >= HttpProtocolVersions.HTTP_1_0:
            out.write(self._header_encode(self._format_status_line(
                a.version,
                a.code,
                a.message,
            )))

            for h in a.headers or []:
                out.write(self._header_encode(self._format_header_line(h)))

            out.write(b'\r\n')

        return out.getvalue()

    def _yield_response_data(self, a: _Response) -> ta.Iterator[bytes]:
        if a.data is None:
            return

        elif isinstance(a.data, bytes):
            yield a.data
            return

        elif isinstance(a.data, HttpHandlerResponseStreamedData):
            yield from a.data.iter

        else:
            raise TypeError(a.data)

    #

    DEFAULT_CONTENT_TYPE = 'text/plain'

    def _preprocess_response(self, resp: _Response) -> _Response:
        nh: ta.List[CoroHttpServer._Header] = []
        kw: ta.Dict[str, ta.Any] = {}

        if resp.get_header('Content-Type') is None:
            nh.append(self._Header('Content-Type', self._default_content_type))

        if resp.data is not None and resp.get_header('Content-Length') is None:
            cl: ta.Optional[int]
            if isinstance(resp.data, bytes):
                cl = len(resp.data)
            elif isinstance(resp.data, HttpHandlerResponseStreamedData):
                cl = resp.data.length
            else:
                raise TypeError(resp.data)
            if cl is not None:
                nh.append(self._Header('Content-Length', str(cl)))

        if nh:
            kw.update(headers=[*(resp.headers or []), *nh])

        if (clh := resp.get_header('Connection')) is not None:
            if self._get_header_close_connection_action(clh):
                kw.update(close_connection=True)

        if not kw:
            return resp
        return dc.replace(resp, **kw)

    #

    @dc.dataclass(frozen=True)
    class Error:
        version: HttpProtocolVersion
        code: http.HTTPStatus
        message: str
        explain: str

        method: ta.Optional[str] = None

    def _build_error(
            self,
            code: ta.Union[http.HTTPStatus, int],
            message: ta.Optional[str] = None,
            explain: ta.Optional[str] = None,
            *,
            version: ta.Optional[HttpProtocolVersion] = None,
            method: ta.Optional[str] = None,
    ) -> Error:
        code = http.HTTPStatus(code)

        try:
            short_msg, long_msg = self._STATUS_RESPONSES[code]
        except KeyError:
            short_msg, long_msg = '???', '???'
        if message is None:
            message = short_msg
        if explain is None:
            explain = long_msg

        if version is None:
            version = self._parser.server_version

        return self.Error(
            version=version,
            code=code,
            message=message,
            explain=explain,

            method=method,
        )

    #

    DEFAULT_ERROR_MESSAGE = textwrap.dedent("""\
        <!DOCTYPE HTML>
        <html lang="en">
            <head>
                <meta charset="utf-8">
                <title>Error response</title>
            </head>
            <body>
                <h1>Error response</h1>
                <p>Error code: %(code)d</p>
                <p>Message: %(message)s.</p>
                <p>Error code explanation: %(code)s - %(explain)s.</p>
            </body>
        </html>
    """)

    DEFAULT_ERROR_CONTENT_TYPE = 'text/html;charset=utf-8'

    def _build_error_response(self, err: Error) -> _Response:
        headers: ta.List[CoroHttpServer._Header] = [
            *self._make_default_headers(),
            self._Header('Connection', 'close'),
        ]

        # Message body is omitted for cases described in:
        #  - RFC7230: 3.3. 1xx, 204(No Content), 304(Not Modified)
        #  - RFC7231: 6.3.6. 205(Reset Content)
        data: ta.Optional[bytes] = None
        if (
                err.code >= http.HTTPStatus.OK and
                err.code not in (
                    http.HTTPStatus.NO_CONTENT,
                    http.HTTPStatus.RESET_CONTENT,
                    http.HTTPStatus.NOT_MODIFIED,
                )
        ):
            # HTML encode to prevent Cross Site Scripting attacks (see bug #1100201)
            content = self._error_message_format.format(
                code=err.code,
                message=html.escape(err.message, quote=False),
                explain=html.escape(err.explain, quote=False),
            )
            body = content.encode('UTF-8', 'replace')

            headers.extend([
                self._Header('Content-Type', self._error_content_type),
                self._Header('Content-Length', str(len(body))),
            ])

            if err.method != 'HEAD' and body:
                data = body

        return self._Response(
            version=err.version,
            code=err.code,
            message=err.message,
            headers=headers,
            data=data,
            close_connection=True,
        )

    #

    class Io(abc.ABC):  # noqa
        pass

    #

    class AnyLogIo(Io):
        pass

    @dc.dataclass(frozen=True)
    class ParsedRequestLogIo(AnyLogIo):
        request: ParsedHttpRequest

    @dc.dataclass(frozen=True)
    class ErrorLogIo(AnyLogIo):
        error: 'CoroHttpServer.Error'

    #

    class AnyReadIo(Io):  # noqa
        pass

    @dc.dataclass(frozen=True)
    class ReadIo(AnyReadIo):
        sz: int

    @dc.dataclass(frozen=True)
    class ReadLineIo(AnyReadIo):
        sz: int

    #

    @dc.dataclass(frozen=True)
    class WriteIo(Io):
        data: bytes

    #

    def coro_handle(self) -> ta.Generator[Io, ta.Optional[bytes], None]:
        return self._coro_run_handler(self._coro_handle_one())

    class Close(Exception):  # noqa
        pass

    def _coro_run_handler(
            self,
            gen: ta.Generator[
                ta.Union[AnyLogIo, AnyReadIo, _Response],
                ta.Optional[bytes],
                None,
            ],
    ) -> ta.Generator[Io, ta.Optional[bytes], None]:
        i: ta.Optional[bytes]
        o: ta.Any = next(gen)
        while True:
            try:
                if isinstance(o, self.AnyLogIo):
                    i = None
                    yield o

                elif isinstance(o, self.AnyReadIo):
                    i = check.isinstance((yield o), bytes)

                elif isinstance(o, self._Response):
                    i = None

                    r = self._preprocess_response(o)
                    hb = self._build_response_head_bytes(r)
                    check.none((yield self.WriteIo(hb)))

                    for b in self._yield_response_data(r):
                        yield self.WriteIo(b)

                    o.close()
                    if o.close_connection:
                        break
                    o = None

                else:
                    raise TypeError(o)  # noqa

                try:
                    o = gen.send(i)
                except self.Close:
                    return
                except StopIteration:
                    break

            except Exception:  # noqa
                if hasattr(o, 'close'):
                    o.close()

                raise

    def _coro_handle_one(self) -> ta.Generator[
        ta.Union[AnyLogIo, AnyReadIo, _Response],
        ta.Optional[bytes],
        None,
    ]:
        # Parse request

        gen = self._parser.coro_parse()
        sz = next(gen)
        while True:
            try:
                line = check.isinstance((yield self.ReadLineIo(sz)), bytes)
                sz = gen.send(line)
            except StopIteration as e:
                parsed = e.value
                break

        if isinstance(parsed, EmptyParsedHttpResult):
            raise self.Close

        if isinstance(parsed, ParseHttpRequestError):
            err = self._build_error(
                parsed.code,
                *([parsed.message] if isinstance(parsed.message, str) else parsed.message),
                version=parsed.version,
            )
            yield self.ErrorLogIo(err)
            yield self._build_error_response(err)
            return

        parsed = check.isinstance(parsed, ParsedHttpRequest)

        # Log

        check.none((yield self.ParsedRequestLogIo(parsed)))

        # Handle CONTINUE

        if parsed.expects_continue:
            # https://bugs.python.org/issue1491
            # https://github.com/python/cpython/commit/0f476d49f8d4aa84210392bf13b59afc67b32b31
            yield self._Response(
                version=parsed.version,
                code=http.HTTPStatus.CONTINUE,
            )

        # Read data

        request_data: ta.Optional[bytes]
        if (cl := parsed.headers.get('Content-Length')) is not None:
            request_data = check.isinstance((yield self.ReadIo(int(cl))), bytes)
        else:
            request_data = None

        # Build request

        handler_request = HttpHandlerRequest(
            client_address=self._client_address,
            method=check.not_none(parsed.method),
            path=parsed.path,
            headers=parsed.headers,
            data=request_data,
        )

        # Build handler response

        try:
            handler_response = self._handler(handler_request)

        except UnsupportedMethodHttpHandlerError:
            err = self._build_error(
                http.HTTPStatus.NOT_IMPLEMENTED,
                f'Unsupported method ({parsed.method!r})',
                version=parsed.version,
                method=parsed.method,
            )
            yield self.ErrorLogIo(err)
            yield self._build_error_response(err)
            return

        try:
            # Build internal response

            response_headers = handler_response.headers or {}
            response_data = handler_response.data

            headers: ta.List[CoroHttpServer._Header] = [
                *self._make_default_headers(),
            ]

            for k, v in response_headers.items():
                headers.append(self._Header(k, v))

            if handler_response.close_connection and 'Connection' not in headers:
                headers.append(self._Header('Connection', 'close'))

            yield self._Response(
                version=parsed.version,
                code=http.HTTPStatus(handler_response.status),
                headers=headers,
                data=response_data,
                close_connection=handler_response.close_connection,
            )

        except Exception:  # noqa
            handler_response.close()

            raise



"""
https://docs.python.org/3/library/bz2.html#bz2.BZ2Compressor
https://docs.python.org/3/library/zlib.html#zlib.decompressobj
https://docs.python.org/3/library/lzma.html#lzma.LZMADecompressor
"""
import abc


##


class CompressorObject(abc.ABC):
    @abc.abstractmethod
    def compress(self, data: bytes) -> bytes:
        """
        Provide data to the compressor object. Returns a chunk of compressed data if possible, or an empty byte string
        otherwise.

        When you have finished providing data to the compressor, call the flush() method to finish the compression
        process.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def flush(self) -> bytes:
        """
        Finish the compression process. Returns the compressed data left in internal buffers.

        The compressor object may not be used after this method has been called.
        """

        raise NotImplementedError


##


class DecompressorObject(abc.ABC):
    @property
    @abc.abstractmethod
    def unused_data(self) -> bytes:
        """
        Data found after the end of the compressed stream.

        If this attribute is accessed before the end of the stream has been reached, its value will be b''.
        """

        raise NotImplementedError

    @property
    @abc.abstractmethod
    def eof(self) -> bool:
        """True if the end-of-stream marker has been reached."""

        raise NotImplementedError

    @abc.abstractmethod
    def decompress(self, data: bytes, *max_length: int) -> bytes:
        """
        Decompress data, returning a bytes object containing the uncompressed data corresponding to at least part of the
        data in string. This data should be concatenated to the output produced by any preceding calls to the
        decompress() method. Some of the input data may be preserved in internal buffers for later processing.

        If the optional parameter max_length is non-zero then the return value will be no longer than max_length.
        """

        raise NotImplementedError


class NeedsInputDecompressorObject(DecompressorObject):
    """
    Used by:
     - bz2.BZ2Decompressor
     - lzma.LZMADecompressor
    """

    @property
    @abc.abstractmethod
    def needs_input(self) -> bool:
        """
        False if the decompress() method can provide more decompressed data before requiring new uncompressed input.
        """

        raise NotImplementedError


class UnconsumedTailDecompressorObject(DecompressorObject):
    """
    Used by:
     - zlib.decompressobj
    """

    @property
    @abc.abstractmethod
    def unconsumed_tail(self) -> bytes:
        """
        A bytes object that contains any data that was not consumed by the last decompress() call because it exceeded
        the limit for the uncompressed data buffer. This data has not yet been seen by the zlib machinery, so you must
        feed it (possibly with further data concatenated to it) back to a subsequent decompress() method call in order
        to get correct output.
        """

        raise NotImplementedError



from .base import Builder
from .binary import BinaryBuilder
from .exprs import ExprBuilder
from .idents import IdentBuilder
from .inserts import InsertBuilder
from .multi import MultiBuilder
from .names import NameBuilder
from .params import ParamBuilder
from .relations import RelationBuilder
from .selects import SelectBuilder
from .stmts import StmtBuilder
from .unary import UnaryBuilder


class StdBuilder(
    InsertBuilder,
    SelectBuilder,
    StmtBuilder,

    RelationBuilder,

    MultiBuilder,
    BinaryBuilder,
    UnaryBuilder,
    ExprBuilder,

    ParamBuilder,

    NameBuilder,
    IdentBuilder,

    Builder,
):
    pass



# ruff: noqa: N802 UP006 UP007
# @omlish-lite
import logging


class ProxyLogFilterer(logging.Filterer):
    def __init__(self, underlying: logging.Filterer) -> None:  # noqa
        self._underlying = underlying

    @property
    def underlying(self) -> logging.Filterer:
        return self._underlying

    @property
    def filters(self):
        return self._underlying.filters

    @filters.setter
    def filters(self, filters):
        self._underlying.filters = filters

    def addFilter(self, filter):  # noqa
        self._underlying.addFilter(filter)

    def removeFilter(self, filter):  # noqa
        self._underlying.removeFilter(filter)

    def filter(self, record):
        return self._underlying.filter(record)


class ProxyLogHandler(ProxyLogFilterer, logging.Handler):
    def __init__(self, underlying: logging.Handler) -> None:  # noqa
        ProxyLogFilterer.__init__(self, underlying)

    _underlying: logging.Handler

    @property
    def underlying(self) -> logging.Handler:
        return self._underlying

    def get_name(self):
        return self._underlying.get_name()

    def set_name(self, name):
        self._underlying.set_name(name)

    @property
    def name(self):
        return self._underlying.name

    @property
    def level(self):
        return self._underlying.level

    @level.setter
    def level(self, level):
        self._underlying.level = level

    @property
    def formatter(self):
        return self._underlying.formatter

    @formatter.setter
    def formatter(self, formatter):
        self._underlying.formatter = formatter

    def createLock(self):
        self._underlying.createLock()

    def acquire(self):
        self._underlying.acquire()

    def release(self):
        self._underlying.release()

    def setLevel(self, level):
        self._underlying.setLevel(level)

    def format(self, record):
        return self._underlying.format(record)

    def emit(self, record):
        self._underlying.emit(record)

    def handle(self, record):
        return self._underlying.handle(record)

    def setFormatter(self, fmt):
        self._underlying.setFormatter(fmt)

    def flush(self):
        self._underlying.flush()

    def close(self):
        self._underlying.close()

    def handleError(self, record):
        self._underlying.handleError(record)



from .errors import InvalidTagContentError


# Mapping for escaping tag values
TAG_VAL_TO_ESCAPE = {
    '\\': '\\\\',
    ';': '\\:',
    ' ': '\\s',
    '\r': '\\r',
    '\n': '\\n',
}


TAG_ESCAPE_CHAR_LOOKUP_TABLE = {i: chr(i) for i in range(256)}  # Most chars escape to themselves

# These are the exceptions
TAG_ESCAPE_CHAR_LOOKUP_TABLE.update({
    ord(':'): ';',
    ord('s'): ' ',
    ord('r'): '\r',
    ord('n'): '\n',
})


def escape_tag_value(in_string: str) -> str:
    for key, val in TAG_VAL_TO_ESCAPE.items():
        in_string = in_string.replace(key, val)
    return in_string


def unescape_tag_value(in_string: str) -> str:
    if '\\' not in in_string:
        return in_string

    buf = []
    remainder = in_string
    while remainder:
        backslash_pos = remainder.find('\\')
        if backslash_pos == -1:
            buf.append(remainder)
            break
        elif backslash_pos == len(remainder) - 1:
            # Trailing backslash, which we strip
            buf.append(remainder[:-1])
            break

        buf.append(remainder[:backslash_pos])
        buf.append(TAG_ESCAPE_CHAR_LOOKUP_TABLE.get(ord(remainder[backslash_pos + 1]), remainder[backslash_pos + 1]))
        remainder = remainder[backslash_pos + 2:]

    return ''.join(buf)


def validate_tag_name(name: str) -> bool:
    if len(name) == 0:
        return False
    if name[0] == '+':
        name = name[1:]
    if len(name) == 0:
        return False
    # Let's err on the side of leniency here; allow -./ (45-47) in any position
    for c in name:  # noqa
        if not (('-' <= c <= '/') or ('0' <= c <= '9') or ('A' <= c <= 'Z') or ('a' <= c <= 'z')):
            return False
    return True


def validate_tag_value(value: str) -> bool:
    rt = value.encode('utf-8', 'ignore').decode('utf-8', 'ignore')
    return value == rt


def parse_tags(raw_tags: str) -> dict[str, str]:
    dct: dict[str, str] = {}

    while raw_tags:
        tag_end = raw_tags.find(';')
        if tag_end == -1:
            tag_pair = raw_tags
            raw_tags = ''
        else:
            tag_pair = raw_tags[:tag_end]
            raw_tags = raw_tags[tag_end + 1:]

        equals_index = tag_pair.find('=')
        if equals_index == -1:
            # Tag with no value
            tag_name, tag_value = tag_pair, ''
        else:
            tag_name, tag_value = tag_pair[:equals_index], tag_pair[equals_index + 1:]

        # "Implementations [...] MUST NOT perform any validation that would reject the message if an invalid tag key
        # name is used."
        if validate_tag_name(tag_name):
            if not validate_tag_value(tag_value):
                raise InvalidTagContentError
            dct[tag_name] = unescape_tag_value(tag_value)

    return dct



# ruff: noqa: UP006 UP007
# @omlish-lite
# Based on bluelet ( https://github.com/sampsyo/bluelet ) by Adrian Sampson, original license:
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
TODO:
 - use fdio
 - wrap coros in Tasks :|
 - (unit)tests lol
 - * subprocesses
 - canceling
 - timeouts
 - task groups
 - gather
 - locks / semaphores / events / etc
 - rename Coro to Bluelet?
 - shutdown
 - ensure resource cleanup
 - run_thread? whatever?

Subprocs:
 - https://github.com/python/cpython/issues/120804 - GH-120804: Remove get_child_watcher and set_child_watcher from
   asyncio
 - https://github.com/python/cpython/pull/17063/files bpo-38692: Add os.pidfd_open
 - clone PidfdChildWatcher + ThreadedChildWatcher
"""
import collections
import dataclasses as dc
import errno
import logging
import select
import sys
import time
import traceback
import types
import typing as ta
import weakref

from .core import BlueletCoro
from .core import BlueletExcInfo
from .core import CoreBlueletEvent
from .core import DelegationBlueletEvent
from .core import ExceptionBlueletEvent
from .core import JoinBlueletEvent
from .core import KillBlueletEvent
from .core import ReturnBlueletEvent
from .core import SleepBlueletEvent
from .core import SpawnBlueletEvent
from .core import ValueBlueletEvent
from .core import _BlueletAwaitableDriver
from .events import BlueletEvent
from .events import BlueletWaitable
from .events import WaitableBlueletEvent


##


class BlueletCoroException(Exception):  # noqa
    def __init__(self, coro: BlueletCoro, exc_info: BlueletExcInfo) -> None:
        super().__init__()
        self.coro = coro
        self.exc_info = exc_info

    @staticmethod
    def _exc_info() -> BlueletExcInfo:
        return sys.exc_info()  # type: ignore

    @staticmethod
    def _reraise(typ: ta.Type[BaseException], exc: BaseException, tb: types.TracebackType) -> ta.NoReturn:  # noqa
        raise exc.with_traceback(tb)

    def reraise(self) -> ta.NoReturn:
        self._reraise(self.exc_info[0], self.exc_info[1], self.exc_info[2])


##


def _bluelet_event_select(
        events: ta.Iterable[BlueletEvent],
        *,
        log: ta.Optional[logging.Logger] = None,
) -> ta.Set[WaitableBlueletEvent]:
    """
    Perform a select() over all the Events provided, returning the ones ready to be fired. Only WaitableEvents
    (including SleepEvents) matter here; all other events are ignored (and thus postponed).
    """

    waitable_to_event: ta.Dict[ta.Tuple[str, BlueletWaitable], WaitableBlueletEvent] = {}
    rlist: ta.List[BlueletWaitable] = []
    wlist: ta.List[BlueletWaitable] = []
    xlist: ta.List[BlueletWaitable] = []
    earliest_wakeup: ta.Optional[float] = None

    # Gather waitables and wakeup times.
    for event in events:
        if isinstance(event, SleepBlueletEvent):
            if not earliest_wakeup:
                earliest_wakeup = event.wakeup_time
            else:
                earliest_wakeup = min(earliest_wakeup, event.wakeup_time)

        elif isinstance(event, WaitableBlueletEvent):
            ew = event.waitables()
            rlist.extend(ew.r)
            wlist.extend(ew.w)
            xlist.extend(ew.x)
            for waitable in ew.r:
                waitable_to_event[('r', waitable)] = event
            for waitable in ew.w:
                waitable_to_event[('w', waitable)] = event
            for waitable in ew.x:
                waitable_to_event[('x', waitable)] = event

    # If we have a any sleeping coros, determine how long to sleep.
    if earliest_wakeup:
        timeout = max(earliest_wakeup - time.time(), 0.)
    else:
        timeout = None

    # Perform select() if we have any waitables.
    if rlist or wlist or xlist:
        if log:
            log.debug('_bluelet_event_select: +select: %r %r %r %r', rlist, wlist, xlist, timeout)
        rready, wready, xready = select.select(rlist, wlist, xlist, timeout)
        if log:
            log.debug('_bluelet_event_select: -select: %r %r %r', rready, wready, xready)

    else:
        rready, wready, xready = [], [], []
        if timeout:
            if log:
                log.debug('_bluelet_event_select: sleep: %r', timeout)
            time.sleep(timeout)

    # Gather ready events corresponding to the ready waitables.
    ready_events: ta.Set[WaitableBlueletEvent] = set()
    for ready in rready:
        ready_events.add(waitable_to_event[('r', ready)])
    for ready in wready:
        ready_events.add(waitable_to_event[('w', ready)])
    for ready in xready:
        ready_events.add(waitable_to_event[('x', ready)])

    # Gather any finished sleeps.
    for event in events:
        if isinstance(event, SleepBlueletEvent) and not event.time_left():
            ready_events.add(event)

    return ready_events


##


class _SuspendedBlueletEvent(CoreBlueletEvent):
    pass


_BLUELET_SUSPENDED = _SuspendedBlueletEvent()  # Special sentinel placeholder for suspended coros.


@dc.dataclass(frozen=True, eq=False)
class _DelegatedBlueletEvent(CoreBlueletEvent):
    """Placeholder indicating that a coro has delegated execution to a different coro."""

    child: BlueletCoro


class _BlueletRunner:
    """
    Schedules a coroutine, running it to completion. This encapsulates the Bluelet scheduler, which the root coroutine
    can add to by spawning new coroutines.
    """

    def __init__(
            self,
            root_coro: BlueletCoro,
            *,
            log: ta.Optional[logging.Logger] = None,
    ) -> None:
        super().__init__()

        self._root_coro = root_coro
        self._log = log

        # The "coros" dictionary keeps track of all the currently-executing and suspended coroutines. It maps
        # coroutines to their currently "blocking" event. The event value may be SUSPENDED if the coroutine is waiting
        # on some other condition: namely, a delegated coroutine or a joined coroutine. In this case, the coroutine
        # should *also* appear as a value in one of the below dictionaries `delegators` or `joiners`.
        self._coros: ta.Dict[BlueletCoro, BlueletEvent] = {self._root_coro: ValueBlueletEvent(None)}

        # Maps child coroutines to delegating parents.
        self._delegators: ta.Dict[BlueletCoro, BlueletCoro] = {}

        # Maps child coroutines to joining (exit-waiting) parents.
        self._joiners: ta.MutableMapping[BlueletCoro, ta.List[BlueletCoro]] = collections.defaultdict(list)

        # History of spawned coroutines for joining of already completed coroutines.
        self._history: ta.MutableMapping[BlueletCoro, ta.Optional[BlueletEvent]] = \
            weakref.WeakKeyDictionary({self._root_coro: None})

    def _complete_coro(self, coro: BlueletCoro, return_value: ta.Any) -> None:
        """
        Remove a coroutine from the scheduling pool, awaking delegators and joiners as necessary and returning the
        specified value to any delegating parent.
        """

        del self._coros[coro]

        # Resume delegator.
        if coro in self._delegators:
            self._coros[self._delegators[coro]] = ValueBlueletEvent(return_value)
            del self._delegators[coro]

        # Resume joiners.
        if coro in self._joiners:
            for parent in self._joiners[coro]:
                self._coros[parent] = ValueBlueletEvent(None)
            del self._joiners[coro]

    def _advance_coro(self, coro: BlueletCoro, value: ta.Any, is_exc: bool = False) -> None:
        """
        After an event is fired, run a given coroutine associated with it in the coros dict until it yields again. If
        the coroutine exits, then the coro is removed from the pool. If the coroutine raises an exception, it is
        reraised in a CoroException. If is_exc is True, then the value must be an exc_info tuple and the exception is
        thrown into the coroutine.
        """

        try:
            if is_exc:
                next_event = coro.throw(value[0])
            else:
                next_event = coro.send(value)

        except StopIteration:
            # Coro is done.
            self._complete_coro(coro, None)

        except BaseException:  # noqa
            # Coro raised some other exception.
            del self._coros[coro]
            # Note: Don't use `raise from` as this should support 3.8.
            raise BlueletCoroException(coro, BlueletCoroException._exc_info())  # noqa

        else:
            if isinstance(next_event, ta.Generator):
                # Automatically invoke sub-coroutines. (Shorthand for explicit bluelet.call().)
                next_event = DelegationBlueletEvent(next_event)

            if isinstance(next_event, types.CoroutineType):  # type: ignore[unreachable]
                next_event = DelegationBlueletEvent(_BlueletAwaitableDriver(next_event)())  # type: ignore[unreachable]

            if not isinstance(next_event, BlueletEvent):
                raise TypeError(next_event)

            self._coros[coro] = next_event

    def _kill_coro(self, coro: BlueletCoro) -> None:
        """Unschedule this coro and its (recursive) delegates."""

        # Collect all coroutines in the delegation stack.
        coros = [coro]
        while isinstance((cur := self._coros[coro]), _DelegatedBlueletEvent):
            coro = cur.child  # noqa
            coros.append(coro)

        # Complete each coroutine from the top to the bottom of the stack.
        for coro in reversed(coros):
            self._complete_coro(coro, None)

    def close(self) -> None:
        # If any coros still remain, kill them.
        for coro in self._coros:
            coro.close()

        self._coros.clear()

    def _handle_core_event(self, coro: BlueletCoro, event: CoreBlueletEvent) -> bool:
        if self._log:
            self._log.debug(f'{self.__class__.__name__}._handle_core_event: %r %r', coro, event)

        if isinstance(event, SpawnBlueletEvent):
            sc = ta.cast(BlueletCoro, event.spawned)  # FIXME
            self._coros[sc] = ValueBlueletEvent(None)  # Spawn.
            self._history[sc] = None  # Record in history.
            self._advance_coro(coro, None)
            return True

        elif isinstance(event, ValueBlueletEvent):
            self._advance_coro(coro, event.value)
            return True

        elif isinstance(event, ExceptionBlueletEvent):
            self._advance_coro(coro, event.exc_info, True)
            return True

        elif isinstance(event, DelegationBlueletEvent):
            self._coros[coro] = _DelegatedBlueletEvent(event.spawned)  # Suspend.
            self._coros[event.spawned] = ValueBlueletEvent(None)  # Spawn.
            self._history[event.spawned] = None  # Record in history.
            self._delegators[event.spawned] = coro
            return True

        elif isinstance(event, ReturnBlueletEvent):
            # Coro is done.
            self._complete_coro(coro, event.value)
            return True

        elif isinstance(event, JoinBlueletEvent):
            if event.child not in self._coros and event.child in self._history:
                self._coros[coro] = ValueBlueletEvent(None)
            else:
                self._coros[coro] = _BLUELET_SUSPENDED  # Suspend.
                self._joiners[event.child].append(coro)
            return True

        elif isinstance(event, KillBlueletEvent):
            self._coros[coro] = ValueBlueletEvent(None)
            self._kill_coro(event.child)
            return True

        elif isinstance(event, (_DelegatedBlueletEvent, _SuspendedBlueletEvent)):
            return False

        else:
            raise TypeError(event)

    def _step(self) -> ta.Optional[BlueletCoroException]:
        if self._log:
            self._log.debug(f'{self.__class__.__name__}._step')  # Noqa

        try:
            # Look for events that can be run immediately. Continue running immediate events until nothing is ready.
            while True:
                have_ready = False
                for coro, event in list(self._coros.items()):
                    if isinstance(event, CoreBlueletEvent) and not isinstance(event, SleepBlueletEvent):
                        have_ready |= self._handle_core_event(coro, event)
                    elif isinstance(event, WaitableBlueletEvent):
                        pass
                    else:
                        raise TypeError(f'Unknown event type: {event}')  # noqa

                # Only start the select when nothing else is ready.
                if not have_ready:
                    break

            # Wait and fire.
            event2coro = {v: k for k, v in self._coros.items()}
            for event in _bluelet_event_select(self._coros.values()):
                # Run the IO operation, but catch socket errors.
                try:
                    value = event.fire()
                except OSError as exc:
                    if isinstance(exc.args, tuple) and exc.args[0] == errno.EPIPE:
                        # Broken pipe. Remote host disconnected.
                        pass
                    elif isinstance(exc.args, tuple) and exc.args[0] == errno.ECONNRESET:
                        # Connection was reset by peer.
                        pass
                    else:
                        traceback.print_exc()
                    # Abort the coroutine.
                    self._coros[event2coro[event]] = ReturnBlueletEvent(None)
                else:
                    self._advance_coro(event2coro[event], value)

        except BlueletCoroException as te:
            if self._log and self._log.isEnabledFor(logging.DEBUG):
                self._log.exception(f'{self.__class__.__name__}._step')

            # Exception raised from inside a coro.
            event = ExceptionBlueletEvent(te.exc_info)
            if te.coro in self._delegators:
                # The coro is a delegate. Raise exception in its delegator.
                self._coros[self._delegators[te.coro]] = event
                del self._delegators[te.coro]
            else:
                # The coro is root-level. Raise in client code.
                return te

        except BaseException:  # noqa
            ei = BlueletCoroException._exc_info()  # noqa

            if self._log and self._log.isEnabledFor(logging.DEBUG):
                self._log.exception(f'{self.__class__.__name__}._step')

            # For instance, KeyboardInterrupt during select(). Raise into root coro and terminate others.
            self._coros = {self._root_coro: ExceptionBlueletEvent(ei)}  # noqa

        return None

    def run(self) -> None:
        # Continue advancing coros until root coro exits.
        exit_ce: BlueletCoroException | None = None
        while self._coros:
            exit_ce = self._step()

        self.close()

        # If we're exiting with an exception, raise it in the client.
        if exit_ce:
            exit_ce.reraise()


##


class _RunnerBlueletApi:
    def run(self, root_coro: BlueletCoro) -> None:
        _BlueletRunner(root_coro).run()



# ruff: noqa: UP006 UP007
# @omlish-lite
"""
Notes:
 - necessarily string-oriented
 - single file, as this is intended to be amalg'd and thus all included anyway

TODO:
 - ConfigDataMapper? to_map -> ConfigMap?
 - nginx ?
 - raw ?
"""
import abc
import collections.abc
import configparser
import dataclasses as dc
import json
import os.path
import typing as ta

from ..formats.ini.sections import IniSectionSettingsMap
from ..formats.ini.sections import extract_ini_sections
from ..formats.ini.sections import render_ini_sections
from ..formats.toml.parser import toml_loads
from ..formats.toml.writer import TomlWriter
from ..lite.check import check
from ..lite.json import json_dumps_pretty
from .types import ConfigMap


T = ta.TypeVar('T')
ConfigDataT = ta.TypeVar('ConfigDataT', bound='ConfigData')


##


@dc.dataclass(frozen=True)
class ConfigData(abc.ABC):  # noqa
    @abc.abstractmethod
    def as_map(self) -> ConfigMap:
        raise NotImplementedError


#


class ConfigLoader(abc.ABC, ta.Generic[ConfigDataT]):
    @property
    def file_exts(self) -> ta.Sequence[str]:
        return ()

    def match_file(self, n: str) -> bool:
        return '.' in n and n.split('.')[-1] in check.not_isinstance(self.file_exts, str)

    #

    def load_file(self, p: str) -> ConfigDataT:
        with open(p) as f:
            return self.load_str(f.read())

    @abc.abstractmethod
    def load_str(self, s: str) -> ConfigDataT:
        raise NotImplementedError


#


class ConfigRenderer(abc.ABC, ta.Generic[ConfigDataT]):
    @property
    @abc.abstractmethod
    def data_cls(self) -> ta.Type[ConfigDataT]:
        raise NotImplementedError

    def match_data(self, d: ConfigDataT) -> bool:
        return isinstance(d, self.data_cls)

    #

    @abc.abstractmethod
    def render(self, d: ConfigDataT) -> str:
        raise NotImplementedError


##


@dc.dataclass(frozen=True)
class ObjConfigData(ConfigData, abc.ABC):
    obj: ta.Any

    def as_map(self) -> ConfigMap:
        return check.isinstance(self.obj, collections.abc.Mapping)


##


@dc.dataclass(frozen=True)
class JsonConfigData(ObjConfigData):
    pass


class JsonConfigLoader(ConfigLoader[JsonConfigData]):
    file_exts = ('json',)

    def load_str(self, s: str) -> JsonConfigData:
        return JsonConfigData(json.loads(s))


class JsonConfigRenderer(ConfigRenderer[JsonConfigData]):
    data_cls = JsonConfigData

    def render(self, d: JsonConfigData) -> str:
        return json_dumps_pretty(d.obj)


##


@dc.dataclass(frozen=True)
class TomlConfigData(ObjConfigData):
    pass


class TomlConfigLoader(ConfigLoader[TomlConfigData]):
    file_exts = ('toml',)

    def load_str(self, s: str) -> TomlConfigData:
        return TomlConfigData(toml_loads(s))


class TomlConfigRenderer(ConfigRenderer[TomlConfigData]):
    data_cls = TomlConfigData

    def render(self, d: TomlConfigData) -> str:
        return TomlWriter.write_str(d.obj)


##


@dc.dataclass(frozen=True)
class YamlConfigData(ObjConfigData):
    pass


class YamlConfigLoader(ConfigLoader[YamlConfigData]):
    file_exts = ('yaml', 'yml')

    def load_str(self, s: str) -> YamlConfigData:
        return YamlConfigData(__import__('yaml').safe_load(s))


class YamlConfigRenderer(ConfigRenderer[YamlConfigData]):
    data_cls = YamlConfigData

    def render(self, d: YamlConfigData) -> str:
        return __import__('yaml').safe_dump(d.obj)


##


@dc.dataclass(frozen=True)
class IniConfigData(ConfigData):
    sections: IniSectionSettingsMap

    def as_map(self) -> ConfigMap:
        return self.sections


class IniConfigLoader(ConfigLoader[IniConfigData]):
    file_exts = ('ini',)

    def load_str(self, s: str) -> IniConfigData:
        cp = configparser.ConfigParser()
        cp.read_string(s)
        return IniConfigData(extract_ini_sections(cp))


class IniConfigRenderer(ConfigRenderer[IniConfigData]):
    data_cls = IniConfigData

    def render(self, d: IniConfigData) -> str:
        return render_ini_sections(d.sections)


##


@dc.dataclass(frozen=True)
class SwitchedConfigFileLoader:
    loaders: ta.Sequence[ConfigLoader]
    default: ta.Optional[ConfigLoader] = None

    def load_file(self, p: str) -> ConfigData:
        n = os.path.basename(p)

        for l in self.loaders:
            if l.match_file(n):
                return l.load_file(p)

        if (d := self.default) is not None:
            return d.load_file(p)

        raise NameError(n)


DEFAULT_CONFIG_LOADERS: ta.Sequence[ConfigLoader] = [
    JsonConfigLoader(),
    TomlConfigLoader(),
    YamlConfigLoader(),
    IniConfigLoader(),
]

DEFAULT_CONFIG_LOADER: ConfigLoader = JsonConfigLoader()

DEFAULT_CONFIG_FILE_LOADER = SwitchedConfigFileLoader(
    loaders=DEFAULT_CONFIG_LOADERS,
    default=DEFAULT_CONFIG_LOADER,
)


##


@dc.dataclass(frozen=True)
class SwitchedConfigRenderer:
    renderers: ta.Sequence[ConfigRenderer]

    def render(self, d: ConfigData) -> str:
        for r in self.renderers:
            if r.match_data(d):
                return r.render(d)
        raise TypeError(d)


DEFAULT_CONFIG_RENDERERS: ta.Sequence[ConfigRenderer] = [
    JsonConfigRenderer(),
    TomlConfigRenderer(),
    YamlConfigRenderer(),
    IniConfigRenderer(),
]

DEFAULT_CONFIG_RENDERER = SwitchedConfigRenderer(DEFAULT_CONFIG_RENDERERS)



# ruff: noqa: UP006 UP007
# @omlish-lite
"""
TODO:
 - default command
 - auto match all underscores to hyphens
 - pre-run, post-run hooks
 - exitstack?
"""
import argparse
import dataclasses as dc
import functools
import inspect
import sys
import typing as ta

from ..lite.check import check
from ..lite.reflect import get_optional_alias_arg
from ..lite.reflect import is_optional_alias


T = ta.TypeVar('T')


ArgparseCmdFn = ta.Callable[[], ta.Optional[int]]  # ta.TypeAlias


##


@dc.dataclass(eq=False)
class ArgparseArg:
    args: ta.Sequence[ta.Any]
    kwargs: ta.Mapping[str, ta.Any]
    dest: ta.Optional[str] = None

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        return getattr(instance.args, self.dest)  # type: ignore


def argparse_arg(*args, **kwargs) -> ArgparseArg:
    return ArgparseArg(args, kwargs)


#


@dc.dataclass(eq=False)
class ArgparseCmd:
    name: str
    fn: ArgparseCmdFn
    args: ta.Sequence[ArgparseArg] = ()  # noqa

    # _: dc.KW_ONLY

    aliases: ta.Optional[ta.Sequence[str]] = None
    parent: ta.Optional['ArgparseCmd'] = None
    accepts_unknown: bool = False

    def __post_init__(self) -> None:
        def check_name(s: str) -> None:
            check.isinstance(s, str)
            check.not_in('_', s)
            check.not_empty(s)
        check_name(self.name)
        check.not_isinstance(self.aliases, str)
        for a in self.aliases or []:
            check_name(a)

        check.arg(callable(self.fn))
        check.arg(all(isinstance(a, ArgparseArg) for a in self.args))
        check.isinstance(self.parent, (ArgparseCmd, type(None)))
        check.isinstance(self.accepts_unknown, bool)

        functools.update_wrapper(self, self.fn)

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        return dc.replace(self, fn=self.fn.__get__(instance, owner))  # noqa

    def __call__(self, *args, **kwargs) -> ta.Optional[int]:
        return self.fn(*args, **kwargs)


def argparse_cmd(
        *args: ArgparseArg,
        name: ta.Optional[str] = None,
        aliases: ta.Optional[ta.Iterable[str]] = None,
        parent: ta.Optional[ArgparseCmd] = None,
        accepts_unknown: bool = False,
) -> ta.Any:  # ta.Callable[[ArgparseCmdFn], ArgparseCmd]:  # FIXME
    for arg in args:
        check.isinstance(arg, ArgparseArg)
    check.isinstance(name, (str, type(None)))
    check.isinstance(parent, (ArgparseCmd, type(None)))
    check.not_isinstance(aliases, str)

    def inner(fn):
        return ArgparseCmd(
            (name if name is not None else fn.__name__).replace('_', '-'),
            fn,
            args,
            aliases=tuple(aliases) if aliases is not None else None,
            parent=parent,
            accepts_unknown=accepts_unknown,
        )

    return inner


##


def _get_argparse_arg_ann_kwargs(ann: ta.Any) -> ta.Mapping[str, ta.Any]:
    if ann is str:
        return {}
    elif ann is int:
        return {'type': int}
    elif ann is bool:
        return {'action': 'store_true'}
    elif ann is list:
        return {'action': 'append'}
    elif is_optional_alias(ann):
        return _get_argparse_arg_ann_kwargs(get_optional_alias_arg(ann))
    else:
        raise TypeError(ann)


class _ArgparseCliAnnotationBox:
    def __init__(self, annotations: ta.Mapping[str, ta.Any]) -> None:
        super().__init__()
        self.__annotations__ = annotations  # type: ignore


class ArgparseCli:
    def __init__(self, argv: ta.Optional[ta.Sequence[str]] = None) -> None:
        super().__init__()

        self._argv = argv if argv is not None else sys.argv[1:]

        self._args, self._unknown_args = self.get_parser().parse_known_args(self._argv)

    #

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        ns = cls.__dict__
        objs = {}
        mro = cls.__mro__[::-1]
        for bns in [bcls.__dict__ for bcls in reversed(mro)] + [ns]:
            bseen = set()  # type: ignore
            for k, v in bns.items():
                if isinstance(v, (ArgparseCmd, ArgparseArg)):
                    check.not_in(v, bseen)
                    bseen.add(v)
                    objs[k] = v
                elif k in objs:
                    del [k]

        #

        anns = ta.get_type_hints(_ArgparseCliAnnotationBox({
            **{k: v for bcls in reversed(mro) for k, v in getattr(bcls, '__annotations__', {}).items()},
            **ns.get('__annotations__', {}),
        }), globalns=ns.get('__globals__', {}))

        #

        if '_parser' in ns:
            parser = check.isinstance(ns['_parser'], argparse.ArgumentParser)
        else:
            parser = argparse.ArgumentParser()
            setattr(cls, '_parser', parser)

        #

        subparsers = parser.add_subparsers()

        for att, obj in objs.items():
            if isinstance(obj, ArgparseCmd):
                if obj.parent is not None:
                    raise NotImplementedError

                for cn in [obj.name, *(obj.aliases or [])]:
                    subparser = subparsers.add_parser(cn)

                    for arg in (obj.args or []):
                        if (
                                len(arg.args) == 1 and
                                isinstance(arg.args[0], str) and
                                not (n := check.isinstance(arg.args[0], str)).startswith('-') and
                                'metavar' not in arg.kwargs
                        ):
                            subparser.add_argument(
                                n.replace('-', '_'),
                                **arg.kwargs,
                                metavar=n,
                            )
                        else:
                            subparser.add_argument(*arg.args, **arg.kwargs)

                    subparser.set_defaults(_cmd=obj)

            elif isinstance(obj, ArgparseArg):
                if att in anns:
                    ann_kwargs = _get_argparse_arg_ann_kwargs(anns[att])
                    obj.kwargs = {**ann_kwargs, **obj.kwargs}

                if not obj.dest:
                    if 'dest' in obj.kwargs:
                        obj.dest = obj.kwargs['dest']
                    else:
                        obj.dest = obj.kwargs['dest'] = att  # type: ignore

                parser.add_argument(*obj.args, **obj.kwargs)

            else:
                raise TypeError(obj)

    #

    _parser: ta.ClassVar[argparse.ArgumentParser]

    @classmethod
    def get_parser(cls) -> argparse.ArgumentParser:
        return cls._parser

    @property
    def argv(self) -> ta.Sequence[str]:
        return self._argv

    @property
    def args(self) -> argparse.Namespace:
        return self._args

    @property
    def unknown_args(self) -> ta.Sequence[str]:
        return self._unknown_args

    #

    def _bind_cli_cmd(self, cmd: ArgparseCmd) -> ta.Callable:
        return cmd.__get__(self, type(self))

    def prepare_cli_run(self) -> ta.Optional[ta.Callable]:
        cmd = getattr(self.args, '_cmd', None)

        if self._unknown_args and not (cmd is not None and cmd.accepts_unknown):
            msg = f'unrecognized arguments: {" ".join(self._unknown_args)}'
            if (parser := self.get_parser()).exit_on_error:  # type: ignore
                parser.error(msg)
            else:
                raise argparse.ArgumentError(None, msg)

        if cmd is None:
            self.get_parser().print_help()
            return None

        return self._bind_cli_cmd(cmd)

    #

    def cli_run(self) -> ta.Optional[int]:
        if (fn := self.prepare_cli_run()) is None:
            return 0

        return fn()

    def cli_run_and_exit(self) -> ta.NoReturn:
        sys.exit(rc if isinstance(rc := self.cli_run(), int) else 0)

    def __call__(self, *, exit: bool = False) -> ta.Optional[int]:  # noqa
        if exit:
            return self.cli_run_and_exit()
        else:
            return self.cli_run()

    #

    async def async_cli_run(
            self,
            *,
            force_async: bool = False,
    ) -> ta.Optional[int]:
        if (fn := self.prepare_cli_run()) is None:
            return 0

        if force_async:
            is_async = True
        else:
            tfn = fn
            if isinstance(tfn, ArgparseCmd):
                tfn = tfn.fn
            is_async = inspect.iscoroutinefunction(tfn)

        if is_async:
            return await fn()
        else:
            return fn()



"""
https://github.com/openshift/docker-distribution/blob/master/docs/spec/manifest-v2-2.md
"""
import typing as ta

from .. import dataclasses as dc
from .. import lang
from .. import marshal as msh


SCHEMA_VERSION = 2


class MediaTypes(lang.Namespace):
    # schema1 (existing manifest format)
    MANIFEST_V1 = 'application/vnd.docker.distribution.manifest.v1+json'

    # New image manifest format (schemaVersion = 2)
    MANIFEST_V2 = 'application/vnd.docker.distribution.manifest.v2+json'

    # Manifest list, aka "fat manifest"
    MANIFEST_LIST = 'application/vnd.docker.distribution.manifest.list.v2+json'

    # Container config JSON
    CONTAINER_CONFIG = 'application/vnd.docker.container.image.v1+json'

    # "Layer", as a gzipped tar
    LAYER = 'application/vnd.docker.image.rootfs.diff.tar.gzip'

    # "Layer", as a gzipped tar that should never be pushed
    LAYER_NEVER_PUSH = 'application/vnd.docker.image.rootfs.foreign.diff.tar.gzip'

    # Plugin config JSON
    PLUGIN_CONFIG = 'application/vnd.docker.plugin.v1+json'


@dc.dataclass(frozen=True, kw_only=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL, unknown_field='x')
@msh.update_fields_metadata(['os_version'], name='os.version')
@msh.update_fields_metadata(['os_features'], name='os.features')
class Platform(lang.Final):
    # The architecture field specifies the CPU architecture, for example amd64 or ppc64le.
    architecture: str

    # The os field specifies the operating system, for example linux or windows.
    os: str

    # The optional os.version field specifies the operating system version, for example 10.0.10586.
    os_version: str | None = None

    # The optional os.features field specifies an array of strings, each listing a required OS feature (for example on
    # Windows win32k).
    os_features: ta.Sequence[ta.Any] | None = None

    # The optional variant field specifies a variant of the CPU, for example v6 to specify a particular CPU variant of
    # the ARM CPU.
    variant: str | None = None

    # The optional features field specifies an array of strings, each listing a required CPU feature (for example sse4
    # or aes).
    features: ta.Sequence[ta.Any] | None = None

    x: ta.Mapping[str, ta.Any] | None = None


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL, unknown_field='x')
class Manifest(lang.Final):
    # The MIME type of the referenced object. This will generally be
    # application/vnd.docker.distribution.manifest.v2+json, but it could also be
    # application/vnd.docker.distribution.manifest.v1+json if the manifest list references a legacy schema-1 manifest.
    media_type: str

    # The size in bytes of the object. This field exists so that a client will have an expected size for the content
    # before validating. If the length of the retrieved content does not match the specified length, the content should
    # not be trusted.
    size: int

    # The digest of the content, as defined by the Registry V2 HTTP API Specification.
    digest: str

    # The platform object describes the platform which the image in the manifest runs on. A full list of valid operating
    # system and architecture values are listed in the Go language documentation for $GOOS and $GOARCH
    platform: Platform | None = None

    x: ta.Mapping[str, ta.Any] | None = None


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL, unknown_field='x')
class ManifestList(lang.Final):
    # This field specifies the image manifest schema version as an integer. This schema uses the version 2.
    schema_version: int
    dc.validate(lambda self: self.schema_version == SCHEMA_VERSION)

    # The MIME type of the manifest list. This should be set to
    # application/vnd.docker.distribution.manifest.list.v2+json.
    media_type: str
    dc.validate(lambda self: self.media_type == MediaTypes.MANIFEST_LIST)

    # The manifests field contains a list of manifests for specific platforms.
    manifests: ta.Sequence[Manifest]

    x: ta.Mapping[str, ta.Any] | None = None


@dc.dataclass(frozen=True)
@msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL, unknown_field='x')
class ImageManifest(lang.Final):
    # This field specifies the image manifest schema version as an integer. This schema uses version 2.
    schema_version: int
    dc.validate(lambda self: self.schema_version == SCHEMA_VERSION)

    # The MIME type of the manifest. This should be set to application/vnd.docker.distribution.manifest.v2+json.
    media_type: str
    dc.validate(lambda self: self.media_type == MediaTypes.MANIFEST_V2)

    @dc.dataclass(frozen=True)
    @msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL, unknown_field='x')
    class Config(lang.Final):
        # The MIME type of the referenced object. This should generally be
        # application/vnd.docker.container.image.v1+json.
        media_type: str

        # The size in bytes of the object. This field exists so that a client will have an expected size for the content
        # before validating. If the length of the retrieved content does not match the specified length, the content
        # should not be trusted.
        size: int

        # The digest of the content, as defined by the Registry V2 HTTP API Specification.
        digest: str

        x: ta.Mapping[str, ta.Any] | None = None

    # The config field references a configuration object for a container, by digest. This configuration item is a JSON
    # blob that the runtime uses to set up the container. This new schema uses a tweaked version of this configuration
    # o allow image content-addressability on the daemon side.
    config: Config | None = None

    @dc.dataclass(frozen=True)
    @msh.update_object_metadata(field_naming=msh.Naming.LOW_CAMEL, unknown_field='x')
    class Layer(lang.Final):
        # The MIME type of the referenced object. This should generally be
        # application/vnd.docker.image.rootfs.diff.tar.gzip. Layers of type
        # application/vnd.docker.image.rootfs.foreign.diff.tar.gzip may be pulled from a remote location but they should
        # never be pushed.
        media_type: str

        # The size in bytes of the object. This field exists so that a client will have an expected size for the content
        # before validating. If the length of the retrieved content does not match the specified length, the content
        # should not be trusted.
        size: int

        # The digest of the content, as defined by the Registry V2 HTTP API Specification.
        digest: str

        # Provides a list of URLs from which the content may be fetched. Content must be verified against the digest and
        # size. This field is optional and uncommon.
        urls: ta.Sequence[str] | None = None

        x: ta.Mapping[str, ta.Any] | None = None

    # The layer list is ordered starting from the base image (opposite order of schema1).
    layers: ta.Sequence[Layer] | None = None

    x: ta.Mapping[str, ta.Any] | None = None



import typing as ta

from .processing import Processor
from .utils import Namespace
from .utils import create_fn
from .utils import set_new_attribute
from .utils import tuple_str


def cmp_fn(
        name: str,
        op: str,
        self_tuple: str,
        other_tuple: str,
        globals: Namespace,  # noqa
) -> ta.Callable:
    return create_fn(
        name,
        ('self', 'other'),
        [
            'if other.__class__ is self.__class__:',
            f' return {self_tuple}{op}{other_tuple}',
            'return NotImplemented',
        ],
        globals=globals,
    )


class OrderProcessor(Processor):
    def _process(self) -> None:
        if not self._info.params.order:
            return

        flds = [f for f in self._info.instance_fields if f.compare]
        self_tuple = tuple_str('self', flds)
        other_tuple = tuple_str('other', flds)
        for name, op in [
            ('__lt__', '<'),
            ('__le__', '<='),
            ('__gt__', '>'),
            ('__ge__', '>='),
        ]:
            if set_new_attribute(self._cls, name, cmp_fn(name, op, self_tuple, other_tuple, globals=self._info.globals)):  # noqa
                raise TypeError(
                    f'Cannot overwrite attribute {name} in class {self._cls.__name__}. '
                    f'Consider using functools.total_ordering',
                )



"""
TODO:
 - bytes-like - bytearray, memoryview
 - FileCodec
 - implement options
"""
import abc
import typing as ta

from .. import check
from .. import dataclasses as dc
from .. import lang
from .. import reflect as rfl
from ..funcs import pairs as fps


I = ta.TypeVar('I')
O = ta.TypeVar('O')


##


class EagerCodec(lang.Abstract, ta.Generic[I, O]):
    @abc.abstractmethod
    def encode(self, i: I) -> O:
        raise NotImplementedError

    @abc.abstractmethod
    def decode(self, o: O) -> I:
        raise NotImplementedError

    def as_pair(self) -> fps.FnPair[I, O]:
        return fps.of(self.encode, self.decode)


class IncrementalCodec(lang.Abstract, ta.Generic[I, O]):
    @abc.abstractmethod
    def encode_incremental(self) -> ta.Generator[O | None, I, None]:
        raise NotImplementedError

    @abc.abstractmethod
    def decode_incremental(self) -> ta.Generator[I | None, O, None]:
        raise NotImplementedError


class ComboCodec(  # noqa
    EagerCodec[I, O],
    IncrementalCodec[I, O],
    lang.Abstract,
    ta.Generic[I, O],
):
    pass


##


def check_codec_name(s: str) -> str:
    check.non_empty_str(s)
    check.not_in('_', s)
    check.equal(s.strip(), s)
    return s


##


@dc.dataclass(frozen=True, kw_only=True)
class Codec:
    name: str = dc.xfield(coerce=check_codec_name)
    aliases: ta.Collection[str] | None = dc.xfield(
        default=None,
        coerce=lang.opt_fn(lambda s: [check_codec_name(a) for a in s]),  # type: ignore
    )

    input: rfl.Type = dc.xfield(coerce=rfl.type_)
    output: rfl.Type = dc.xfield(coerce=rfl.type_)

    options: type | None = None

    new: ta.Callable[..., EagerCodec]
    new_incremental: ta.Callable[..., IncrementalCodec] | None = None


##


@dc.dataclass(frozen=True, kw_only=True)
class LazyLoadedCodec:
    mod_name: str
    attr_name: str
    name: str
    aliases: ta.Collection[str] | None = None

    @classmethod
    def new(
            cls,
            mod_name: str,
            attr_name: str,
            codec: Codec,
    ) -> 'LazyLoadedCodec':
        return cls(
            mod_name=mod_name,
            attr_name=attr_name,
            name=codec.name,
            aliases=codec.aliases,
        )



import typing as ta

from .. import cached
from .. import check
from .. import dataclasses as dc
from .. import lang
from .. import reflect as rfl
from ..funcs import match as mfs
from .base import MarshalContext
from .base import Marshaler
from .base import MarshalerFactory
from .base import UnmarshalContext
from .base import Unmarshaler
from .base import UnmarshalerFactory
from .polymorphism import Impls
from .polymorphism import TypeTagging
from .polymorphism import WrapperTypeTagging
from .polymorphism import make_polymorphism_marshaler
from .polymorphism import make_polymorphism_unmarshaler
from .values import Value


##


class MatchUnionMarshaler(Marshaler):
    mmf: mfs.MultiMatchFn[[UnmarshalContext, Value], ta.Any]

    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        try:
            m = self.mmf.match(ctx, o)
        except mfs.AmbiguousMatchesError:
            raise ValueError(o)  # noqa
        return m.fn(ctx, o)


class MatchUnionUnmarshaler(Unmarshaler):
    mmf: mfs.MultiMatchFn[[UnmarshalContext, Value], ta.Any]

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        try:
            m = self.mmf.match(ctx, v)
        except mfs.AmbiguousMatchesError:
            raise ValueError(v)  # noqa
        return m.fn(ctx, v)


##


PRIMITIVE_UNION_TYPES: tuple[type, ...] = (
    float,
    int,
    str,
)


#


@dc.dataclass(frozen=True)
class PrimitiveUnionMarshaler(Marshaler):
    tys: ta.Sequence[type]

    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        if type(o) not in self.tys:
            raise TypeError(o)
        return o


@dc.dataclass(frozen=True)
class PrimitiveUnionMarshalerFactory(MarshalerFactory):
    tys: ta.Sequence[type] = PRIMITIVE_UNION_TYPES

    def guard(self, ctx: MarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, rfl.Union) and all(a in self.tys for a in rty.args)

    def fn(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        args = check.isinstance(rty, rfl.Union).args
        return PrimitiveUnionMarshaler([t for t in self.tys if t in args])


#


@dc.dataclass(frozen=True)
class PrimitiveUnionUnmarshaler(Unmarshaler):
    tys: ta.Sequence[type]

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        if type(v) not in self.tys:
            raise TypeError(v)
        return v


@dc.dataclass(frozen=True)
class PrimitiveUnionUnmarshalerFactory(UnmarshalerFactory):
    tys: ta.Sequence[type] = PRIMITIVE_UNION_TYPES

    def guard(self, ctx: UnmarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, rfl.Union) and all(a in self.tys for a in rty.args)

    def fn(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        args = check.isinstance(rty, rfl.Union).args
        return PrimitiveUnionUnmarshaler([t for t in self.tys if t in args])


#


PRIMITIVE_UNION_MARSHALER_FACTORY = PrimitiveUnionMarshalerFactory()
PRIMITIVE_UNION_UNMARSHALER_FACTORY = PrimitiveUnionUnmarshalerFactory()


##


@dc.dataclass(frozen=True)
class _BasePolymorphismUnionFactory(lang.Abstract):
    impls: Impls
    tt: TypeTagging = WrapperTypeTagging()
    allow_partial: bool = dc.field(default=False, kw_only=True)

    @cached.property
    @dc.init
    def rtys(self) -> frozenset[rfl.Type]:
        return frozenset(i.ty for i in self.impls)

    def guard(self, ctx: MarshalContext | UnmarshalContext, rty: rfl.Type) -> bool:
        if not isinstance(rty, rfl.Union):
            return False
        if self.allow_partial:
            return not (rty.args - self.rtys)
        else:
            return rty.args == self.rtys

    def get_impls(self, rty: rfl.Type) -> Impls:
        uty = check.isinstance(rty, rfl.Union)
        return Impls([self.impls.by_ty[check.isinstance(a, type)] for a in uty.args])


@dc.dataclass(frozen=True)
class PolymorphismUnionMarshalerFactory(_BasePolymorphismUnionFactory, MarshalerFactory):
    def fn(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        return make_polymorphism_marshaler(self.get_impls(rty), self.tt, ctx)


@dc.dataclass(frozen=True)
class PolymorphismUnionUnmarshalerFactory(_BasePolymorphismUnionFactory, UnmarshalerFactory):
    def fn(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        return make_polymorphism_unmarshaler(self.get_impls(rty), self.tt, ctx)



import abc
import dataclasses as dc
import typing as ta

from .. import lang


BootstrapConfigT = ta.TypeVar('BootstrapConfigT', bound='Bootstrap.Config')


##


class Bootstrap(abc.ABC, lang.PackageSealed, ta.Generic[BootstrapConfigT]):
    @dc.dataclass(frozen=True)
    class Config(abc.ABC):  # noqa
        pass

    def __init__(self, config: BootstrapConfigT) -> None:
        super().__init__()
        self._config = config

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)
        if not cls.__name__.endswith('Bootstrap'):
            raise NameError(cls)
        if abc.ABC not in cls.__bases__ and not issubclass(cls.__dict__['Config'], Bootstrap.Config):
            raise TypeError(cls)


class SimpleBootstrap(Bootstrap[BootstrapConfigT], abc.ABC):
    @abc.abstractmethod
    def run(self) -> None:
        raise NotImplementedError


class ContextBootstrap(Bootstrap[BootstrapConfigT], abc.ABC):
    @abc.abstractmethod
    def enter(self) -> ta.ContextManager[None]:
        raise NotImplementedError



import sys
import time
import typing as ta


T = ta.TypeVar('T')


##


class ProgressBar:
    """
    TODO:
     - ProgressBarRenderer
     - right-justify
     - ProgressBarGroup
     - animate
    """

    def __init__(
            self,
            total: int | None = None,
            *,
            length: int = 40,
            interval: float = .2,
            start_time: float | None = None,
            out: ta.TextIO | None = None,
    ) -> None:
        super().__init__()

        self._total = total
        self._length = length
        self._interval = interval
        if start_time is None:
            start_time = time.time()
        self._start_time = start_time
        if out is None:
            out = sys.stdout
        self._out = out

        self._i = 0
        self._elapsed = 0.
        self._last_print = 0.

    def render(
            self,
            *,
            complete: bool = False,
    ) -> str:
        iter_per_sec = self._i / self._elapsed if self._elapsed > 0 else 0

        if self._total is not None:
            remaining = (self._total - self._i) / iter_per_sec if iter_per_sec > 0 else 0
            done = int(self._length * self._i / self._total)

            bar = f'[{"█" * done}{"." * (self._length - done)}]'
            info_parts = [
                f'{self._i}/{self._total}',
                f'{iter_per_sec:.2f} it/s',
                f'{self._elapsed:.2f}s elapsed',
                f'{remaining:.2f}s left',
            ]

        else:
            bar = f'[{("█" if complete else "?") * self._length}]'
            info_parts = [
                f'{self._i}',
                f'{iter_per_sec:.2f} it/s',
                f'{self._elapsed:.2f}s elapsed',
            ]

        info = ' | '.join(info_parts)
        return f'{bar} {info}'

    def print(
            self,
            *,
            now: float | None = None,
            **kwargs: ta.Any,
    ) -> None:
        if now is None:
            now = time.time()

        line = self.render(**kwargs)
        self._out.write(f'\033[2K\033[G{line}')
        self._out.flush()

        self._last_print = now

    def update(
            self,
            n: int = 1,
            *,
            now: float | None = None,
            silent: bool = False,
    ) -> None:
        if now is None:
            now = time.time()

        self._i += n
        self._elapsed = now - self._start_time

        if not silent:
            if now - self._last_print >= self._interval:
                self.print(now=now)


def progress_bar(
        seq: ta.Iterable[T],
        *,
        no_tty_check: bool = False,
        total: int | None = None,
        out: ta.TextIO | None = None,
        **kwargs: ta.Any,
) -> ta.Generator[T, None, None]:
    if out is None:
        out = sys.stdout

    if not no_tty_check and not out.isatty():
        yield from seq
        return

    if total is None:
        if isinstance(seq, ta.Sized):
            total = len(seq)

    pb = ProgressBar(
        total=total,
        out=out,
        **kwargs,
    )

    for item in seq:
        pb.update()
        yield item

    pb.print(complete=True)
    out.write('\n')



import abc
import os
import signal
import sys
import time
import typing as ta

from .. import check


##


class Deathpact(abc.ABC):
    @abc.abstractmethod
    def poll(self) -> None:
        raise NotImplementedError


class NopDeathpact(Deathpact):
    def poll(self) -> None:
        pass


##


class BaseDeathpact(Deathpact, abc.ABC):
    def __init__(
            self,
            *,
            interval_s: float = .5,
            signal: int | None = signal.SIGTERM,
            output: ta.Literal['stdout', 'stderr'] | None = 'stderr',
            on_die: ta.Callable[[], None] | None = None,
    ) -> None:
        super().__init__()

        self._interval_s = interval_s
        self._signal = signal
        self._output = output
        self._on_die = on_die

        self._last_check_t: float | None = None

    def _print(self, msg: str) -> None:
        match self._output:
            case 'stdout':
                f = sys.stdout
            case 'stderr':
                f = sys.stderr
            case _:
                return
        print(f'{self} pid={os.getpid()}: {msg}', file=f)

    def die(self) -> None:
        self._print('Triggered! Process terminating!')

        if self._on_die is not None:
            self._on_die()

        if self._signal is not None:
            os.kill(os.getpid(), self._signal)

        sys.exit(1)

    @abc.abstractmethod
    def should_die(self) -> bool:
        raise NotImplementedError

    def maybe_die(self) -> None:
        if self.should_die():
            self.die()

    def poll(self) -> None:
        if self._last_check_t is None or (time.monotonic() - self._last_check_t) >= self._interval_s:
            self.maybe_die()
            self._last_check_t = time.monotonic()


##


class PipeDeathpact(BaseDeathpact):
    def __init__(self, **kwargs: ta.Any) -> None:
        super().__init__(**kwargs)

        self._rfd: int | None = None
        self._wfd: int | None = None

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(rfd={self._rfd}, wfd={self._wfd})'

    @property
    def pass_fd(self) -> int:
        return check.not_none(self._rfd)

    def __enter__(self) -> ta.Self:
        check.none(self._rfd)
        check.none(self._wfd)

        self._rfd, self._wfd = os.pipe()

        os.set_inheritable(self._rfd, True)
        os.set_blocking(self._rfd, False)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._rfd is not None:
            os.close(check.not_none(self._wfd))
            self._wfd = None

            os.close(self._rfd)
            self._rfd = None

    def should_die(self) -> bool:
        try:
            buf = os.read(check.not_none(self._rfd), 1)
        except BlockingIOError:
            return False

        if buf:
            self._print(f'Read data from pipe! This should not happen! Process state corrupt!')
            self.die()

        return True



# ruff: noqa: UP006 UP007
# @omlish-lite
import abc
import logging
import sys
import typing as ta

from .callers import LoggingCaller


LogLevel = int  # ta.TypeAlias


##


class Logging(ta.Protocol):
    def isEnabledFor(self, level: LogLevel) -> bool:  # noqa
        ...

    def getEffectiveLevel(self) -> LogLevel:  # noqa
        ...

    #

    def debug(self, msg: str, *args, **kwargs) -> None:
        ...

    def info(self, msg: str, *args, **kwargs) -> None:
        ...

    def warning(self, msg: str, *args, **kwargs) -> None:
        ...

    def error(self, msg: str, *args, **kwargs) -> None:
        ...

    def exception(self, msg: str, *args, exc_info=True, **kwargs) -> None:
        ...

    def critical(self, msg: str, *args, **kwargs) -> None:
        ...

    def log(self, level: LogLevel, msg: str, *args, **kwargs) -> None:
        ...


##


class AbstractLogging(abc.ABC):
    @ta.final
    def isEnabledFor(self, level: LogLevel) -> bool:  # noqa
        return self.is_enabled_for(level)

    def is_enabled_for(self, level: LogLevel) -> bool:  # noqa
        return level >= self.getEffectiveLevel()

    @ta.final
    def getEffectiveLevel(self) -> LogLevel:  # noqa
        return self.get_effective_level()

    @abc.abstractmethod
    def get_effective_level(self) -> LogLevel:  # noqa
        raise NotImplementedError

    #

    def debug(self, msg: str, *args, **kwargs) -> None:
        if self.is_enabled_for(logging.DEBUG):
            self.log(logging.DEBUG, msg, args, **kwargs)

    def info(self, msg: str, *args, **kwargs) -> None:
        if self.is_enabled_for(logging.INFO):
            self.log(logging.INFO, msg, args, **kwargs)

    def warning(self, msg: str, *args, **kwargs) -> None:
        if self.is_enabled_for(logging.WARNING):
            self.log(logging.WARNING, msg, args, **kwargs)

    def error(self, msg: str, *args, **kwargs) -> None:
        if self.is_enabled_for(logging.ERROR):
            self.log(logging.ERROR, msg, args, **kwargs)

    def exception(self, msg: str, *args, exc_info=True, **kwargs) -> None:
        self.error(msg, *args, exc_info=exc_info, **kwargs)

    def critical(self, msg: str, *args, **kwargs) -> None:
        if self.is_enabled_for(logging.CRITICAL):
            self.log(logging.CRITICAL, msg, args, **kwargs)

    def log(self, level: LogLevel, msg: str, *args, **kwargs) -> None:
        if not isinstance(level, int):
            raise TypeError('Level must be an integer.')
        if self.is_enabled_for(level):
            self._log(level, msg, args, **kwargs)

    @abc.abstractmethod
    def _log(
            self,
            level: int,
            msg: str,
            args: ta.Any,
            *,
            exc_info: ta.Any = None,
            extra: ta.Any = None,
            stack_info: bool = False,
    ) -> None:
        raise NotImplementedError


##


class NopLogging(AbstractLogging):
    def get_effective_level(self) -> LogLevel:
        return logging.CRITICAL + 1

    def _log(self, *args: ta.Any, **kwargs: ta.Any) -> None:
        pass


##


class StdlibLogging(AbstractLogging):
    def __init__(self, underlying: logging.Logger) -> None:
        super().__init__()

        if not isinstance(underlying, logging.Logger):
            raise TypeError(underlying)

        self._underlying = underlying

    #

    def is_enabled_for(self, level: int) -> bool:  # noqa
        return self._underlying.isEnabledFor(level)

    def get_effective_level(self) -> int:  # noqa
        return self._underlying.getEffectiveLevel()

    #

    def _log(
            self,
            level: int,
            msg: str,
            args: ta.Any,
            *,
            exc_info: ta.Any = None,
            extra: ta.Any = None,
            stack_info: bool = False,
    ) -> None:
        caller = LoggingCaller.find(stack_info)

        if exc_info:
            if isinstance(exc_info, BaseException):
                exc_info = (type(exc_info), exc_info, exc_info.__traceback__)
            elif not isinstance(exc_info, tuple):
                exc_info = sys.exc_info()

        record = self._underlying.makeRecord(
            name=self._underlying.name,
            level=level,
            fn=caller.filename,
            lno=caller.lineno,
            msg=msg,
            args=args,
            exc_info=exc_info,
            func=caller.func,
            extra=extra,
            sinfo=caller.sinfo,
        )

        self._underlying.handle(record)



import argparse
import dataclasses as dc
import sys
import typing as ta

from ...formats import json
from .ast import Node
from .exceptions import ArityError
from .exceptions import JmespathTypeError
from .exceptions import JmespathValueError
from .exceptions import ParseError
from .exceptions import UnknownFunctionError
from .parser import compile  # noqa
from .parser import search
from .visitor import node_type


def _ast_to_json(o: ta.Any) -> ta.Any:
    if isinstance(o, json.SCALAR_TYPES):
        return o
    elif isinstance(o, Node):
        return {node_type(o): {f.name: _ast_to_json(getattr(o, f.name)) for f in dc.fields(o)}}
    elif isinstance(o, (list, tuple)):
        return [_ast_to_json(e) for e in o]
    else:
        raise TypeError(o)


def _main() -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument('expression')
    parser.add_argument('-f', '--filename')
    parser.add_argument('-c', '--compact', action='store_true')
    parser.add_argument('--ast', action='store_true')
    args = parser.parse_args()

    if args.compact:
        json_dumps = json.dumps_compact
    else:
        json_dumps = json.dumps_pretty

    expression = args.expression
    if args.ast:
        expression = compile(args.expression)
        print(json_dumps(_ast_to_json(expression.parsed)))
        return 0

    if args.filename:
        with open(args.filename) as f:
            data = json.load(f)
    else:
        data = sys.stdin.read()
        data = json.loads(data)

    try:
        print(json_dumps(search(expression, data), ensure_ascii=False))
        return 0

    except ArityError as e:
        print(f'invalid-arity: {e}', file=sys.stderr)
        return 1

    except JmespathTypeError as e:
        print(f'invalid-type: {e}', file=sys.stderr)
        return 1

    except JmespathValueError as e:
        print(f'invalid-value: {e}', file=sys.stderr)
        return 1

    except UnknownFunctionError as e:
        print(f'unknown-function: {e}', file=sys.stderr)
        return 1

    except ParseError as e:
        print(f'syntax-error: {e}', file=sys.stderr)
        return 1


if __name__ == '__main__':
    sys.exit(_main())



from ... import lang
from .fields import field_assign
from .params import get_field_extras
from .processing import Processor
from .utils import create_fn
from .utils import set_new_attribute


class OverridesProcessor(Processor):
    def _process(self) -> None:
        for f in self._info.instance_fields:
            fx = get_field_extras(f)
            if not (fx.override or self._info.params_extras.override):
                continue

            if self._info.params.slots:
                raise TypeError

            self_name = '__dataclass_self__' if 'self' in self._info.fields else 'self'

            getter = create_fn(
                f.name,
                (self_name,),
                [f'return {self_name}.__dict__[{f.name!r}]'],
                globals=self._info.globals,
                return_type=lang.just(f.type),
            )
            prop = property(getter)

            if not self._info.params.frozen:
                setter = create_fn(
                    f.name,
                    (self_name, f'{f.name}: __dataclass_type_{f.name}__'),
                    [
                        field_assign(
                            self._info.params.frozen,
                            f.name,
                            f.name,
                            self_name,
                            True,
                        ),
                    ],
                    globals=self._info.globals,
                    locals={f'__dataclass_type_{f.name}__': f.type},
                    return_type=lang.just(None),
                )
                prop = prop.setter(setter)

            set_new_attribute(
                self._cls,
                f.name,
                prop,
            )



# ruff: noqa: UP006 UP007
# @omlish-lite
import contextlib
import os
import shutil
import tempfile
import typing as ta

from .files import unlink_if_exists


def make_temp_file(**kwargs: ta.Any) -> str:
    file_fd, file = tempfile.mkstemp(**kwargs)
    os.close(file_fd)
    return file


@contextlib.contextmanager
def temp_file_context(**kwargs: ta.Any) -> ta.Iterator[str]:
    path = make_temp_file(**kwargs)
    try:
        yield path
    finally:
        unlink_if_exists(path)


@contextlib.contextmanager
def temp_dir_context(
        root_dir: ta.Optional[str] = None,
        **kwargs: ta.Any,
) -> ta.Iterator[str]:
    path = tempfile.mkdtemp(dir=root_dir, **kwargs)
    try:
        yield path
    finally:
        shutil.rmtree(path, ignore_errors=True)


@contextlib.contextmanager
def temp_named_file_context(
        root_dir: ta.Optional[str] = None,
        cleanup: bool = True,
        **kwargs: ta.Any,
) -> ta.Iterator[tempfile._TemporaryFileWrapper]:  # noqa
    with tempfile.NamedTemporaryFile(dir=root_dir, delete=False, **kwargs) as f:
        try:
            yield f
        finally:
            if cleanup:
                shutil.rmtree(f.name, ignore_errors=True)



"""
TODO:
 - @property dfs, bfs
"""
import functools
import typing as ta
import weakref

from .. import cached
from .. import check
from .. import collections as col
from .. import lang
from ..algorithm import all as alg


T = ta.TypeVar('T')
NodeT = ta.TypeVar('NodeT')
NodeWalker = ta.Callable[[NodeT], ta.Iterable[NodeT]]
NodeGenerator = ta.Generator[NodeT, None, None]


class NodeError(ta.Generic[NodeT], Exception):
    def __init__(self, node: NodeT, msg: str, *args, **kwargs) -> None:
        super().__init__(msg, *args, **kwargs)  # noqa
        self._node = node

    @property
    def node(self) -> NodeT:
        return self._node


class DuplicateNodeError(NodeError[NodeT]):
    def __init__(self, node: NodeT, *args, **kwargs) -> None:
        super().__init__(node, f'Duplicate node: {node!r}', *args, **kwargs)


class UnknownNodeError(NodeError[NodeT]):
    def __init__(self, node: NodeT, *args, **kwargs) -> None:
        super().__init__(node, f'Unknown node: {node!r}', *args, **kwargs)


class BasicTreeAnalysis(ta.Generic[NodeT]):

    def __init__(
            self,
            root: NodeT,
            walker: NodeWalker[NodeT],
            *,
            identity: bool = False,
    ) -> None:
        super().__init__()

        self._root = root
        self._walker = walker
        self._identity = identity

        self._set_fac: ta.Callable[..., ta.MutableSet[NodeT]] = col.IdentitySet if identity else set
        self._dict_fac: ta.Callable[..., ta.MutableMapping[NodeT, ta.Any]] = col.IdentityKeyDict if identity else dict
        self._idx_seq_fac: ta.Callable[..., col.IndexedSeq[NodeT]] = functools.partial(col.IndexedSeq, identity=identity)  # type: ignore  # noqa

        def walk(cur: NodeT, parent: NodeT | None) -> None:
            check.not_none(cur)
            if cur in node_set:
                raise DuplicateNodeError(cur)

            nodes.append(cur)
            node_set.add(cur)
            if parent is None:
                check.state(cur is root)
            elif parent not in node_set:
                raise UnknownNodeError(parent)

            parents_by_node[cur] = parent

            children_by_node[cur] = children = list(walker(cur))
            child_sets_by_node[cur] = self._set_fac(children)
            for child in children:
                walk(child, cur)

        nodes: list[NodeT] = []
        node_set: ta.MutableSet[NodeT] = self._set_fac()
        children_by_node: ta.MutableMapping[NodeT | None, ta.Sequence[NodeT]] = self._dict_fac()
        child_sets_by_node: ta.MutableMapping[NodeT | None, ta.AbstractSet[NodeT]] = self._dict_fac()
        parents_by_node: ta.MutableMapping[NodeT, NodeT | None] = self._dict_fac()

        children_by_node[None] = [root]
        child_sets_by_node[None] = self._set_fac([root])

        walk(root, None)

        self._nodes = self._idx_seq_fac(nodes)
        self._node_set: ta.AbstractSet[NodeT] = node_set
        self._children_by_node: ta.Mapping[NodeT | None, col.IndexedSeq[NodeT]] = self._dict_fac(
            [(n, self._idx_seq_fac(cs)) for n, cs in children_by_node.items()])
        self._child_sets_by_node: ta.Mapping[NodeT | None, ta.AbstractSet[NodeT]] = child_sets_by_node
        self._parents_by_node: ta.Mapping[NodeT, NodeT | None] = parents_by_node

    @property
    def root(self) -> NodeT:
        return self._root

    @property
    def nodes(self) -> col.IndexedSeq[NodeT]:
        return self._nodes

    @property
    def walker(self) -> NodeWalker[NodeT]:
        return self._walker

    @property
    def identity(self) -> bool:
        return self._identity

    @property
    def node_set(self) -> ta.AbstractSet[NodeT]:
        return self._node_set

    @property
    def children_by_node(self) -> ta.Mapping[NodeT | None, col.IndexedSeq[NodeT]]:
        return self._children_by_node

    @property
    def child_sets_by_node(self) -> ta.Mapping[NodeT | None, ta.AbstractSet[NodeT]]:
        return self._child_sets_by_node

    @property
    def parents_by_node(self) -> ta.Mapping[NodeT, NodeT | None]:
        return self._parents_by_node

    @classmethod
    def from_parents(
            cls,
            src: ta.Mapping[NodeT, NodeT | None] | ta.Iterable[tuple[NodeT, NodeT | None]],
            *,
            identity: bool = False,
            **kwargs,
    ) -> 'BasicTreeAnalysis[NodeT]':
        pairs: ta.Sequence[tuple[NodeT, NodeT]]
        if isinstance(src, ta.Mapping):
            pairs = list(src.items())  # type: ignore
        elif isinstance(src, ta.Iterable):
            pairs = list(src)  # type: ignore
        else:
            raise TypeError(src)

        pairs = [(check.not_none(n), p) for n, p in pairs]

        root = check.single([n for n, p in pairs if p is None])  # noqa

        children_by_node: ta.MutableMapping[NodeT, ta.MutableSequence[NodeT]] = col.IdentityKeyDict() if identity else {}  # noqa
        for n, _ in pairs:
            children_by_node[n] = []
        for n, p in pairs:
            if p is not None:
                children_by_node[p].append(n)

        return cls(
            root,
            children_by_node.__getitem__,
            identity=identity,
            **kwargs,
        )

    @classmethod
    def from_children(
            cls,
            src: ta.Mapping[NodeT, ta.Iterable[NodeT]] | ta.Iterable[tuple[NodeT, ta.Iterable[NodeT]]],
            *,
            identity: bool = False,
            **kwargs,
    ) -> 'BasicTreeAnalysis[NodeT]':
        pairs: ta.Sequence[tuple[NodeT, ta.Sequence[NodeT]]]
        if isinstance(src, ta.Mapping):
            pairs = list(src.items())  # type: ignore
        elif isinstance(src, ta.Iterable):
            pairs = list(src)  # type: ignore
        else:
            raise TypeError(src)

        pairs = [(check.not_none(n), [check.not_none(c) for c in cs]) for n, cs in pairs]

        children_by_node: ta.MutableMapping[NodeT, ta.Sequence[NodeT]] = col.IdentityKeyDict() if identity else {}
        parents_by_node: ta.MutableMapping[NodeT, NodeT] = col.IdentityKeyDict() if identity else {}
        for n, cs in pairs:
            check.not_in(n, children_by_node)
            children_by_node[n] = cs
            for c in cs:
                check.not_in(c, parents_by_node)
                parents_by_node[c] = n

        e: ta.Any
        d: ta.Any
        if identity:
            e, d = id, col.make_map(((id(n), n) for n, _ in pairs), strict=True)
        else:
            e, d = lang.identity, lang.identity
        tsd = {e(n): {e(p)} for n, p in parents_by_node.items()}
        ts = list(alg.mut_toposort(tsd))
        root = d(check.single(ts[0]))

        return cls(
            root,
            children_by_node.__getitem__,
            identity=identity,
            **kwargs,
        )

    @cached.property
    def _node_sets_by_type(self) -> ta.MutableMapping[type, ta.AbstractSet[NodeT]]:
        return weakref.WeakKeyDictionary()

    def get_node_type_set(self, ty: type[T]) -> ta.AbstractSet[T]:
        try:
            return self._node_sets_by_type[ty]  # type: ignore
        except KeyError:
            ret = self._node_sets_by_type[ty] = self._set_fac(n for n in self.nodes if isinstance(n, ty))
            return ret  # type: ignore

    def iter_ancestors(self, node: NodeT) -> NodeGenerator[NodeT]:
        cur: NodeT | None = node
        while True:
            cur = self.parents_by_node.get(cur)  # type: ignore
            if cur is None:
                break
            yield cur

    def get_lineage(self, node: NodeT) -> col.IndexedSeq[NodeT]:
        return self._idx_seq_fac(reversed([node, *self.iter_ancestors(node)]))

    def get_first_parent_of_type(self, node: NodeT, ty: type[T]) -> T | None:
        for cur in self.iter_ancestors(node):
            if isinstance(cur, ty):
                return cur
        return None

    @cached.property
    def depths_by_node(self) -> ta.Mapping[NodeT, int]:
        def rec(n, d):
            dct[n] = d
            for c in self._children_by_node[n]:
                rec(c, d + 1)
        dct: ta.MutableMapping[NodeT, int] = self._dict_fac()
        rec(self._root, 0)
        return dct



# ruff: noqa: UP007
# @omlish-lite
import contextlib
import os
import threading
import typing as ta
import weakref

from ...lite.check import check
from .pidfile import Pidfile


##


class _PidfileManager:
    """
    Manager for controlled inheritance of Pidfiles across forks in the presence of multiple threads. There is of course
    no safe or correct way to mix the use of fork and multiple active threads, and one should never write code which
    does so, but in the Real World one may still find oneself in such a situation outside of their control (such as when
    running under Pycharm's debugger which forces the use of forked multiprocessing).

    Not implemented as an instantiated class as there is no way to unregister at_fork listeners, and because Pidfiles
    may be pickled and there must be no possibility of accidentally unpickling and instantiating a new instance of the
    manager.
    """

    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError

    _lock: ta.ClassVar[threading.Lock] = threading.Lock()
    _installed: ta.ClassVar[bool] = False
    _pidfile_threads: ta.ClassVar[ta.MutableMapping[Pidfile, threading.Thread]] = weakref.WeakKeyDictionary()

    @classmethod
    def _before_fork(cls) -> None:
        cls._lock.acquire()

    @classmethod
    def _after_fork_in_parent(cls) -> None:
        cls._lock.release()

    @classmethod
    def _after_fork_in_child(cls) -> None:
        th = threading.current_thread()
        for pf, pf_th in list(cls._pidfile_threads.items()):
            if pf_th is not th:
                pf.close()
                del cls._pidfile_threads[pf]

        cls._lock.release()

    #

    @classmethod
    def _install(cls) -> None:
        check.state(not cls._installed)

        os.register_at_fork(
            before=cls._before_fork,
            after_in_parent=cls._after_fork_in_parent,
            after_in_child=cls._after_fork_in_child,
        )

        cls._installed = True

    @classmethod
    def install(cls) -> bool:
        with cls._lock:
            if cls._installed:
                return False

            cls._install()
            return True

    @classmethod
    @contextlib.contextmanager
    def inheritable_pidfile_context(
            cls,
            path: str,
            *,
            inheritable: bool = True,
            **kwargs: ta.Any,
    ) -> ta.Iterator[Pidfile]:
        """
        A contextmanager for creating and managing a Pidfile which will only be inherited by forks of the calling /
        creating thread.
        """

        check.arg(inheritable)

        cls.install()

        pf = Pidfile(
            path,
            inheritable=False,
            **kwargs,
        )

        with cls._lock:
            cls._pidfile_threads[pf] = threading.current_thread()
        try:

            with pf:
                os.set_inheritable(check.not_none(pf.fileno()), True)
                yield pf

        finally:
            with cls._lock:
                del cls._pidfile_threads[pf]


def open_inheritable_pidfile(path: str, **kwargs: ta.Any) -> ta.ContextManager[Pidfile]:
    return _PidfileManager.inheritable_pidfile_context(path, **kwargs)  # noqa



"""
TODO:
 - Namer: ta.TypeAlias = ta.Callable[[str], str] ?
  - this interface is ~intentionally~ limited, but custom overrides would be useful
"""
import enum

from .. import lang
from .base import Option


class Naming(Option, enum.Enum):
    SNAKE = 'snake'
    CAMEL = 'camel'
    LOW_CAMEL = 'low_camel'


def translate_name(n: str, e: Naming) -> str:
    if e is Naming.SNAKE:
        return lang.snake_case(n)
    if e is Naming.CAMEL:
        return lang.camel_case(n)
    if e is Naming.LOW_CAMEL:
        r = lang.camel_case(n)
        return (r[0].lower() + r[1:]) if r else r
    raise ValueError(e)



# ruff: noqa: UP006 UP007
# @omlish-lite
import contextlib
import os
import typing as ta


def touch(self, mode: int = 0o666, exist_ok: bool = True) -> None:
    if exist_ok:
        # First try to bump modification time
        # Implementation note: GNU touch uses the UTIME_NOW option of the utimensat() / futimens() functions.
        try:
            os.utime(self, None)
        except OSError:
            pass
        else:
            return
    flags = os.O_CREAT | os.O_WRONLY
    if not exist_ok:
        flags |= os.O_EXCL
    fd = os.open(self, flags, mode)
    os.close(fd)


def unlink_if_exists(path: str) -> None:
    try:
        os.unlink(path)
    except FileNotFoundError:
        pass


@contextlib.contextmanager
def unlinking_if_exists(path: str) -> ta.Iterator[None]:
    try:
        yield
    finally:
        unlink_if_exists(path)



# ruff: noqa: N802 N803
import typing as ta

from .. import check
from . import runtime as antlr4
from .errors import SilentRaisingErrorListener


LexerT = ta.TypeVar('LexerT', bound=antlr4.Lexer)
ParserT = ta.TypeVar('ParserT', bound=antlr4.Parser)


def is_eof_context(ctx: antlr4.ParserRuleContext) -> bool:
    return ctx.getChildCount() == 1 and ctx.getChild(0).getSymbol().type == antlr4.Token.EOF


class StandardParseTreeVisitor(antlr4.ParseTreeVisitor):
    def visit(self, ctx: antlr4.ParserRuleContext):
        check.isinstance(ctx, antlr4.ParserRuleContext)
        return ctx.accept(self)

    def aggregateResult(self, aggregate, nextResult):  # noqa
        if aggregate is not None:
            check.none(nextResult)
            return aggregate
        else:
            check.none(aggregate)
            return nextResult


def make_parser(
        buf: str,
        lexer_cls: type[LexerT],
        parser_cls: type[ParserT],
        *,
        silent_errors: bool = False,
) -> ParserT:
    lexer = lexer_cls(antlr4.InputStream(buf))
    if silent_errors:
        lexer.removeErrorListeners()
        lexer.addErrorListener(SilentRaisingErrorListener())

    stream = antlr4.CommonTokenStream(lexer)
    stream.fill()

    parser = parser_cls(stream)
    if silent_errors:
        parser.removeErrorListeners()
        parser.addErrorListener(SilentRaisingErrorListener())

    return parser



# ruff: noqa: UP006 UP007
# @omlish-lite
"""
https://man7.org/linux/man-pages/man8/lsof.8.html

Included in the process set are fields that identify the command, the process group IDentification (PGID) number, the
task (thread) ID (TID), and the user ID (UID) number or login name.
"""
import dataclasses as dc
import enum
import typing as ta

from ..lite.check import check
from ..lite.dataclasses import dataclass_repr_omit_falsey
from ..lite.marshal import OBJ_MARSHALER_OMIT_IF_NONE
from ..subprocesses.run import SubprocessRun
from ..subprocesses.run import SubprocessRunnable
from ..subprocesses.run import SubprocessRunOutput


##


# FIXME: ??
# https://unix.stackexchange.com/a/86011
class LsofItemMode(enum.Enum):
    SOLARIS_NFS_LOCK = 'N'
    READ_LOCK_PARTIAL = 'r'
    READ_LOCK_FULL = 'R'
    WRITE_LOCK_PARTIAL = 'w'
    WRITE_LOCK_FULL = 'W'
    READ_WRITE_LOCK = 'u'
    UNKNOWN_LOCK_TYPE = 'U'
    XENIX_LOCK_PARTIAL = 'x'
    XENIX_LOCK_FULL = 'X'


@dc.dataclass(frozen=True)
class _LsofFieldMeta:
    prefix: str
    process: bool = False
    variadic: bool = False

    @classmethod
    def make(cls, *args, **kwargs):
        return {
            cls: cls(*args, **kwargs),
            OBJ_MARSHALER_OMIT_IF_NONE: True,
        }


@dc.dataclass(frozen=True)
class LsofItem:
    __repr__ = dataclass_repr_omit_falsey

    class _PREFIX:
        def __new__(cls, *args, **kwargs):  # noqa
            raise TypeError

    class _PROCESS:
        def __new__(cls, *args, **kwargs):  # noqa
            raise TypeError

    pid: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('p', process=True))
    pgid: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('g', process=True))
    ppid: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('R', process=True))
    cmd: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('c', process=True))

    uid: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('u', process=True))
    login_name: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('L', process=True))

    # FD is the File Descriptor number of the file or:
    #   cwd  current working directory;
    #   Lnn  library references (AIX);
    #   ctty character tty;
    #   DEL  deleted file;
    #   err  FD information error (see NAME column);
    #   fp.  Fileport (Darwin);
    #   jld  jail directory (FreeBSD);
    #   ltx  shared library text (code and data);
    #   Mxx  hex memory-mapped type number xx.
    #   m86  DOS Merge mapped file;
    #   mem  memory-mapped file;
    #   mmap memory-mapped device;
    #   NOFD for a Linux /proc/<PID>/fd directory that can't be opened -- the directory path appears in the NAME column,
    #     followed by an error message;
    #   pd   parent directory;
    #   Rnn  unknown pregion number (HP-UX);
    #   rtd  root directory;
    #   twd  per task current working directory;
    #   txt  program text (code and data);
    #   v86  VP/ix mapped file;
    #
    # FD is followed by one of these characters, describing the mode under which the file is open:
    #   r for read access;
    #   w for write access;
    #   u for read and write access;
    #   space if mode unknown and no lock
    #   character follows;
    #   `-' if mode unknown and lock
    #   character follows.
    #
    # The mode character is followed by one of these lock characters, describing the type of lock applied to the file:
    #   N for a Solaris NFS lock of unknown type;
    #   r for read lock on part of the file;
    #   R for a read lock on the entire file;
    #   w for a write lock on part of the file;
    #   W for a write lock on the entire file;
    #   u for a read and write lock of any length;
    #   U for a lock of unknown type;
    #   x for an SCO OpenServer Xenix lock on part of the file;
    #   X for an SCO OpenServer Xenix lock on the entire file;
    #   space if there is no lock.
    #
    fd: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('f'))

    inode: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('i'))
    name: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('n'))

    # r = read; w = write; u = read/write
    access: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('a'))

    # r/R = read; w/W = write; u = read/write
    lock: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('l'))

    file_flags: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('G'))
    file_type: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('t'))
    file_size: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('s'))
    file_offset: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('o'))  # as 0t<dec> or 0x<hex>

    device_character_code: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('d'))
    device_number: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('D'))  # as 0x<hex>
    raw_device_number: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('r'))  # as 0x<hex>

    share_count: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('C'))
    link_count: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('k'))

    stream_info: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('S'))
    protocol_name: ta.Optional[str] = dc.field(default=None, metadata=_LsofFieldMeta.make('P'))

    tcp_tpi_info: ta.Optional[ta.Sequence[str]] = dc.field(
        default=None,
        metadata=_LsofFieldMeta.make('T', variadic=True),
    )

    # 'K':  'task_id',  # unknown field
    # 'M':  'task_command_name',  # unknown field

    _FIELDS_BY_PREFIX: ta.ClassVar[ta.Mapping[str, dc.Field]]
    _DEFAULT_PREFIXES: ta.ClassVar[str]

    @classmethod
    def from_prefixes(cls, dct: ta.Mapping[str, ta.Any]) -> 'LsofItem':
        kw: ta.Dict[str, ta.Any] = {
            fld.name: val
            for pfx, val in dct.items()
            if (fld := cls._FIELDS_BY_PREFIX.get(pfx)) is not None
        }

        return cls(**kw)

    @classmethod
    def from_prefix_lines(cls, lines: ta.Iterable[str]) -> ta.List['LsofItem']:
        proc_dct: ta.Dict[str, ta.Any] = {}
        file_dct: ta.Dict[str, ta.Any] = {}
        out: ta.List[LsofItem] = []

        def emit() -> None:
            if file_dct:
                out.append(cls.from_prefixes({**proc_dct, **file_dct}))

        for line in lines:
            pfx, val = line[0], line[1:]
            try:
                fld = cls._FIELDS_BY_PREFIX[pfx]
            except KeyError:
                continue
            meta: _LsofFieldMeta = fld.metadata[_LsofFieldMeta]

            if pfx == 'p':
                emit()
                proc_dct = {}
                file_dct = {}
            elif pfx == 'f':
                emit()
                file_dct = {}

            if meta.process:
                dct = proc_dct
            else:
                dct = file_dct

            if meta.variadic:
                dct.setdefault(pfx, []).append(val)
            else:
                if pfx in dct:
                    raise KeyError(pfx)
                dct[pfx] = val

        emit()

        return out


LsofItem._FIELDS_BY_PREFIX = {  # noqa
    meta.prefix: fld
    for fld in dc.fields(LsofItem)  # noqa
    if (meta := fld.metadata.get(_LsofFieldMeta)) is not None  # noqa
}

LsofItem._DEFAULT_PREFIXES = ''.join(LsofItem._FIELDS_BY_PREFIX)  # noqa


##


@dc.dataclass(frozen=True)
class \
        LsofCommand(SubprocessRunnable[ta.List[LsofItem]]):
    pid: ta.Optional[int] = None
    file: ta.Optional[str] = None

    prefixes: ta.Optional[ta.Sequence[str]] = None

    def make_run(self) -> SubprocessRun:
        if (prefixes := self.prefixes) is None:
            prefixes = LsofItem._DEFAULT_PREFIXES  # noqa

        return SubprocessRun.of(
            'lsof',
            '-F', ''.join(prefixes),
            *(['-p', str(self.pid)] if self.pid is not None else []),
            *([self.file] if self.file is not None else []),

            stdout='pipe',
            stderr='devnull',
            check=True,
        )

    def handle_run_output(self, output: SubprocessRunOutput) -> ta.List[LsofItem]:
        lines = [s for l in check.not_none(output.stdout).decode().splitlines() if (s := l.strip())]
        return LsofItem.from_prefix_lines(lines)


if __name__ == '__main__':
    def _main() -> None:
        argparse = __import__('argparse')
        parser = argparse.ArgumentParser()
        parser.add_argument('--pid', '-p', type=int)
        parser.add_argument('file', nargs='?')
        args = parser.parse_args()

        importlib = __import__('importlib')
        subprocesses = importlib.import_module('..subprocesses.sync', package=__package__).subprocesses
        items = LsofCommand(
            pid=args.pid,
            file=args.file,
        ).run(subprocesses)

        json = __import__('json')
        marshal_obj = importlib.import_module('..lite.marshal', package=__package__).marshal_obj
        print(json.dumps(marshal_obj(items), indent=2))

    _main()



# ruff: noqa: I001
from .asyncio import (  # noqa
    asyncio_once as once,
    asyncio_wait_concurrent as wait_concurrent,
    drain_asyncio_tasks as drain_tasks,
    draining_asyncio_tasks as draining_tasks,
)



"""
TODO:
 - https://github.com/aristocratos/btop/blob/00c90884c328eb3e44a0ab094e833161092aae9c/src/linux/btop_collect.cpp#L443

==

Repls
 - python - https://docs.python.org/3/tutorial/interpreter.html
 - ipython - https://github.com/ipython/ipython/
 - ptpython - https://github.com/prompt-toolkit/ptpython
 - bpython - https://github.com/bpython/bpython

Debuggers
 - pdb - https://docs.python.org/3/library/pdb.html
 - ipdb - https://github.com/gotcha/ipdb
 - pdub - https://github.com/inducer/pudb/

CPU Profilers
 - cProfile - https://docs.python.org/3/library/profile.html
 - pyinstrument - https://github.com/joerick/pyinstrument
 - py-pspy - https://github.com/benfred/py-spy
 - austin-dist - https://github.com/P403n1x87/austin
 - yappi - https://github.com/sumerc/yappi

Memory Profilers
 - tracemalloc - https://docs.python.org/3/library/tracemalloc.html
 - guppy3 - https://github.com/zhuyifei1999/guppy3/
 - objgraph - https://github.com/mgedmin/objgraph
 - memory-profiler - https://github.com/pythonprofilers/memory_profiler
 - memray - https://github.com/bloomberg/memray
 - pympler - https://github.com/pympler/pympler
"""



import types
import typing as ta


class ValidationError(Exception):
    pass


def _hands_off_repr(obj: ta.Any) -> str:
    return f'{obj.__class__.__qualname__}@{hex(id(obj))[2:]}'


def _fn_repr(fn: ta.Callable) -> str:
    if (co := getattr(fn, '__code__', None)) is None or not isinstance(co, types.CodeType):
        return repr(fn)

    if not (co_filename := co.co_filename):
        return repr(fn)

    return f'{fn!r} ({co_filename}:{co.co_firstlineno})'


class FieldValidationError(ValidationError):
    def __init__(
            self,
            obj: ta.Any,
            field: str,
            fn: ta.Callable,
            value: ta.Any,
    ) -> None:
        super().__init__(
            f'{self.__class__.__name__} '
            f'for field {field!r} '
            f'on object {_hands_off_repr(obj)} '
            f'in validator {_fn_repr(fn)} '
            f'with value {value!r}',
        )

        self.obj = obj
        self.field = field
        self.fn = fn
        self.value = value

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({", ".join([
            f"obj={_hands_off_repr(self.obj)}",
            f"field={self.field!r}",
            f"fn={_fn_repr(self.fn)}",
            f"value={self.value!r}",
        ])})'



# ruff: noqa: UP006 UP007
# @omlish-lite
import dataclasses as dc
import os
import typing as ta


@dc.dataclass(frozen=True)
class DockerPortRelay:
    docker_port: int
    host_port: int

    name: ta.Optional[str] = None

    DEFAULT_HOST_NAME: ta.ClassVar[str] = 'host.docker.internal'
    host_name: str = DEFAULT_HOST_NAME

    DEFAULT_INTERMEDIATE_PORT: ta.ClassVar[int] = 5000
    intermediate_port: int = DEFAULT_INTERMEDIATE_PORT

    DEFAULT_IMAGE: ta.ClassVar[str] = 'alpine/socat'
    image: str = DEFAULT_IMAGE

    def socat_args(self) -> ta.List[str]:
        return [
            '-d',
            f'TCP-LISTEN:{self.intermediate_port},fork,reuseaddr',
            f'TCP:{self.host_name}:{self.host_port}',
        ]

    def run_args(self) -> ta.List[str]:
        if (name := self.name) is None:
            name = f'docker_port_relay-{os.getpid()}'

        return [
            '--name', name,
            '--rm',
            '-p', f'{self.docker_port}:{self.intermediate_port}',
            self.image,
            *self.socat_args(),
        ]

    def run_cmd(self) -> ta.List[str]:
        return [
            'docker',
            'run',
            '-i',
            *self.run_args(),
        ]
